import logging
import tiktoken
from llama_index.core.callbacks.schema import CBEventType
from llama_index.core.callbacks import CallbackManager, TokenCountingHandler
from llama_index.core import Settings
from typing import Tuple, List, Any, Optional
from lavague.core.utilities.pricing_util import get_pricing_data
from lavague.core.world_model import WorldModel
from lavague.core.action_engine import ActionEngine
from lavague.core.base_engine import ActionResult
from lavague.core.world_model import WORLD_MODEL_PROMPT_TEMPLATE

# used by gpt-4* models
DEFAULT_TOKENIZER = "o200k_base"


class TokenCounter:
    """
    Token counter for each of MM LLM and embedding models.
    It relies on TokenCountingHandler from llama_index.
    Only one token counter can be used at a time.
    """

    def __init__(
        self,
        log=False,
    ):
        self.pricing_data = get_pricing_data()
        default_tokenizer = tiktoken.get_encoding(DEFAULT_TOKENIZER).encode

        mm_llm_token_counter = TokenCountingHandler(
            tokenizer=default_tokenizer,
            event_starts_to_ignore=[CBEventType.EMBEDDING],
            event_ends_to_ignore=[CBEventType.EMBEDDING],
        )
        embedding_token_counter = TokenCountingHandler(
            tokenizer=default_tokenizer,
            event_starts_to_ignore=[CBEventType.LLM],
            event_ends_to_ignore=[CBEventType.LLM],
        )
        Settings.callback_manager = CallbackManager(
            [mm_llm_token_counter, embedding_token_counter]
        )
        self.mm_llm_token_counter = mm_llm_token_counter
        self.embedding_token_counter = embedding_token_counter

        if log:
            logging_print = logging.getLogger(__name__)
            logging_print.setLevel(logging.INFO)
            format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
            ch = logging.StreamHandler()
            ch.setLevel(logging.INFO)
            ch.setFormatter(format)
            logging_print.addHandler(ch)
            logging_print.propagate = False
            self.logging_print = logging_print
        else:
            self.logging_print = None

    def get_token_count(self, event_type):
        return self.token_counter.get_total_tokens(event_type)

    def process_token_usage(
        self,
        world_model: WorldModel,
        action_engine: ActionEngine,
        result_to_update: Optional[ActionResult] = None,
    ) -> Tuple[dict, dict]:
        """Computes token counts and costs for the current step"""

        if (
            self.embedding_token_counter is not None
            and self.mm_llm_token_counter is not None
        ):
            mm_llm_name = world_model.get_mm_llm_name()
            llm_name = action_engine.get_llm_name()
            embedding_name = action_engine.get_embedding_name()
            # deduplicate and split llm events into world model and action engine events
            deduplicated_events = deduplicate_events(
                self.mm_llm_token_counter.llm_token_counts
            )
            world_model_events, action_engine_events = triage_events(
                deduplicated_events
            )

            # compute llm token counts for each models
            WM_input_tokens, WM_output_tokens, WM_total_tokens = self.count_tokens(
                world_model_events, mm_llm_name
            )
            AE_input_tokens, AE_output_tokens, AE_total_tokens = self.count_tokens(
                action_engine_events, llm_name
            )

            # compute embedding token counts
            embedding_total_tokens = (
                self.embedding_token_counter.total_embedding_token_count
            )

            total_step_tokens = (
                WM_total_tokens + AE_total_tokens + embedding_total_tokens
            )

            token_counts = {
                "world_model_input_tokens": WM_input_tokens,
                "world_model_output_tokens": WM_output_tokens,
                "action_engine_input_tokens": AE_input_tokens,
                "action_engine_output_tokens": AE_output_tokens,
                "total_world_model_tokens": WM_total_tokens,
                "total_action_engine_tokens": AE_total_tokens,
                "total_llm_tokens": WM_total_tokens + AE_total_tokens,
                "total_embedding_tokens": embedding_total_tokens,
                "total_step_tokens": total_step_tokens,
            }

            # compute llm and embedding token costs
            WM_input_cost, WM_output_cost, WM_total_cost = self.calculate_llm_pricing(
                WM_input_tokens, WM_output_tokens, mm_llm_name
            )
            AE_input_cost, AE_output_cost, AE_total_cost = self.calculate_llm_pricing(
                AE_input_tokens,
                AE_output_tokens,
                llm_name,
            )

            total_embedding_cost = self.calculate_embedding_pricing(
                embedding_total_tokens,
                embedding_name,
            )
            total_step_cost = WM_total_cost + AE_total_cost + total_embedding_cost

            token_costs = {
                "world_model_input_cost": WM_input_cost,
                "world_model_output_cost": WM_output_cost,
                "action_engine_input_cost": AE_input_cost,
                "action_engine_output_cost": AE_output_cost,
                "total_world_model_cost": WM_total_cost,
                "total_action_engine_cost": AE_total_cost,
                "total_llm_cost": WM_total_cost + AE_total_cost,
                "total_embedding_cost": total_embedding_cost,
                "total_step_cost": total_step_cost,
            }

            if result_to_update is not None:
                result_to_update.total_estimated_tokens += total_step_tokens
                result_to_update.total_estimated_cost += total_step_cost

            # reset counters
            self.mm_llm_token_counter.reset_counts()
            self.embedding_token_counter.reset_counts()
        else:
            # when token counters are not initialized, we return 0s
            token_counts = {
                "world_model_input_tokens": 0,
                "world_model_output_tokens": 0,
                "action_engine_input_tokens": 0,
                "action_engine_output_tokens": 0,
                "total_world_model_tokens": 0,
                "total_action_engine_tokens": 0,
                "total_llm_tokens": 0,
                "total_embedding_tokens": 0,
                "total_step_tokens": 0,
            }
            token_costs = {
                "world_model_input_cost": 0,
                "world_model_output_cost": 0,
                "action_engine_input_cost": 0,
                "action_engine_output_cost": 0,
                "total_world_model_cost": 0,
                "total_action_engine_cost": 0,
                "total_llm_cost": 0,
                "total_embedding_cost": 0,
                "total_step_cost": 0,
            }

        if self.logging_print:
            counts = [
                f"{key}: {value}" for key, value in token_counts.items() if value > 0
            ]
            if len(counts) > 0:
                self.logging_print.info("Token consumption:\n" + "\n".join(counts))

        return token_counts, token_costs

    def calculate_llm_pricing(
        self, input_token_count: int, output_token_count: int, model: str
    ) -> Tuple[int, int, int]:
        """Computes token costs for LLM according to the pricing data available in pricing_config.yaml"""
        input_pricing = self.pricing_data.get(model, {model: {"input_tokens": 0}}).get(
            "input_tokens", 0
        )
        output_pricing = self.pricing_data.get(
            model, {model: {"output_tokens": 0}}
        ).get("output_tokens", 0)
        token_ratio = self.pricing_data.get(model, {model: {"token_ratio": 1}}).get(
            "token_ratio", 1
        )

        llm_input_token_cost = (input_token_count * input_pricing) / token_ratio
        llm_output_token_cost = (output_token_count * output_pricing) / token_ratio

        llm_cost = llm_input_token_cost + llm_output_token_cost

        return llm_input_token_cost, llm_output_token_cost, llm_cost

    def calculate_embedding_pricing(self, token_count: int, model: str) -> int:
        """Computes token costs for Embedding according to the pricing data available in pricing_config.yaml"""
        embedding_pricing = self.pricing_data.get(
            model, {model: {"input_tokens": 0}}
        ).get("input_tokens", 0)
        token_ratio = self.pricing_data.get(model, {model: {"token_ratio": 1}}).get(
            "token_ratio", 1
        )
        total_embedding_cost = (token_count * embedding_pricing) / token_ratio

        return total_embedding_cost

    def count_tokens(self, events: List[Any], model) -> Tuple[int, int, int]:
        """Helper function to sum token counts for a list of events that belong together"""
        # we use a multiplier to approximate tokens consumed by gemini models based on our default tokenizer

        token_multiplier = self.pricing_data.get(
            model, {model: {"token_multiplier": 1}}
        ).get("token_multiplier", 1)

        input_tokens = (
            sum(event.prompt_token_count for event in events) * token_multiplier
        )
        output_tokens = (
            sum(event.completion_token_count for event in events) * token_multiplier
        )
        total_tokens = (
            sum(event.total_token_count for event in events) * token_multiplier
        )
        return input_tokens, output_tokens, total_tokens


def deduplicate_events(events: List[Any]) -> List[Any]:
    """Helper function to deduplicate events if they are logged twice (temporary fix for #444)"""

    def are_events_equal(event1: Any, event2: Any) -> bool:
        """Helper function to compare two events by their attributes"""
        attributes_to_compare = [
            "prompt",
            "completion",
            "completion_token_count",
            "prompt_token_count",
            "total_token_count",
        ]
        return all(
            getattr(event1, attr) == getattr(event2, attr)
            for attr in attributes_to_compare
        )

    deduplicated = []
    for event in events:
        if not any(
            are_events_equal(existing_event, event) for existing_event in deduplicated
        ):
            deduplicated.append(event)
    return deduplicated


def triage_events(events: List[Any]) -> Tuple[List[Any], List[Any]]:
    """Split events into world model and action engine events based on prompt start"""
    world_model_events = []
    other_events = []

    world_model_prompt_start = WORLD_MODEL_PROMPT_TEMPLATE.template.strip().split("\n")[
        0
    ]

    for event in events:
        if event.prompt.strip().startswith(world_model_prompt_start):
            world_model_events.append(event)
        else:
            other_events.append(event)

    return world_model_events, other_events



"""TODO: We used to have a version of this that was quite flexible and could allow for others to provide their capabilities and the Python engine could use them.
However, we have decided to simplify the Python engine to only use the LLM and Embedding from the context for the moment as the Python Engine we want is a bit complex.
We will revisit this in the future to be able to have both:
- Capabilities that can be provided by others
- State management that can allow a complex agentic workflow"""

from io import BytesIO
import time
from lavague.core.context import Context, get_default_context
from lavague.core.utilities.web_utils import display_screenshot
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.embeddings import BaseEmbedding
from lavague.core.base_driver import BaseDriver

import trafilatura
from llama_index.core import Document, VectorStoreIndex
from lavague.core.logger import AgentLogger
from lavague.core.base_engine import BaseEngine, ActionResult
from PIL import Image
from typing import Callable


class PythonEngine(BaseEngine):
    llm: BaseLLM
    embedding: BaseEmbedding
    driver: BaseDriver

    def __init__(
        self,
        driver: BaseDriver,
        llm: BaseLLM = None,
        embedding: BaseEmbedding = None,
        logger: AgentLogger = None,
        clean_html: Callable[[str], str] = None,
    ):
        if llm is None:
            llm = get_default_context().llm
        if embedding is None:
            embedding = get_default_context().embedding
        if clean_html is None:
            self.clean_html = trafilatura.extract

        self.llm = llm
        self.embedding = embedding
        self.driver = driver
        self.logger = logger
        self.display = False

    @classmethod
    def from_context(
        cls,
        context: Context,
    ):
        return cls(context.llm, context.embedding)

    def execute_instruction(self, instruction: str) -> ActionResult:
        logger = self.logger

        html = self.driver.get_html()
        start = time.time()
        llm = self.llm
        embedding = self.embedding

        if self.display:
            try:
                screenshot = self.driver.get_screenshot_as_png()
                screenshot = BytesIO(screenshot)
                screenshot = Image.open(screenshot)
                display_screenshot(screenshot)
            except:
                pass

        page_content = self.clean_html(html)

        documents = [Document(text=page_content)]

        index = VectorStoreIndex.from_documents(documents, embed_model=embedding)
        query_engine = index.as_query_engine(llm=llm)

        output = query_engine.query(instruction).response
        end = time.time()
        action_time = end - start

        success = True

        if logger:
            log = {
                "engine": "Python Engine",
                "instruction": instruction,
                "engine_log": {
                    "action_time": action_time,
                },
                "success": success,
                "output": output,
                "code": "",  # TODO add code of Python engine. Issue is that it might be harder to make it re runnable as the pipeline is more complex than just navigation.
            }

            logger.add_log(log)

        return ActionResult(
            instruction=instruction, code="", success=success, output=output
        )

    def set_display(self, display: bool):
        self.display = display



from __future__ import annotations
from typing import List, Optional, Tuple
from abc import ABC, abstractmethod
from bs4 import BeautifulSoup, NavigableString
from llama_index.retrievers.bm25 import BM25Retriever
from llama_index.core import Document, VectorStoreIndex, QueryBundle
from llama_index.core.schema import NodeWithScore, TextNode
from langchain.text_splitter import RecursiveCharacterTextSplitter
from llama_index.core.node_parser import LangchainNodeParser
from llama_index.core.embeddings import BaseEmbedding
from lavague.core.base_driver import BaseDriver, PossibleInteractionsByXpath
from lavague.core.utilities.format_utils import clean_html
import re
import ast


def get_default_retriever(
    driver: BaseDriver, embedding: Optional[BaseEmbedding] = None
) -> BaseHtmlRetriever:
    return RetrieversPipeline(
        InteractiveXPathRetriever(driver),
        FromXPathNodesExpansionRetriever(),
        SemanticRetriever(embedding=embedding),
    )


class BaseHtmlRetriever(ABC):
    @abstractmethod
    def retrieve(
        self, query: QueryBundle, html_nodes: List[str], viewport_only=True
    ) -> List[str]:
        """
        This method must be implemented by the child retriever
        """
        pass


class RetrieversPipeline(BaseHtmlRetriever):
    """Executor for retrievers pipeline"""

    retrievers: Tuple[BaseHtmlRetriever]

    def __init__(self, *retrievers: BaseHtmlRetriever):
        self.retrievers = retrievers

    def retrieve(
        self, query: QueryBundle, html_nodes: List[str], viewport_only=True
    ) -> List[str]:
        for retriever in self.retrievers:
            html_nodes = retriever.retrieve(query, html_nodes, viewport_only)
        return html_nodes


class BM25HtmlRetriever(BaseHtmlRetriever):
    """Mainly for benchmarks, do not use it as the performances are not up to par with the other retrievers"""

    def __init__(self, top_k=3) -> None:
        self.top_k = top_k

    def retrieve(self, query: QueryBundle, html_chunks: List[str]) -> List[str]:
        html = clean_html(merge_html_chunks(html_chunks))
        cleaned_html = clean_html(html)

        splitter = LangchainNodeParser(
            lc_splitter=RecursiveCharacterTextSplitter.from_language(
                language="html",
            )
        )
        nodes = splitter.get_nodes_from_documents([Document(text=cleaned_html)])

        retriever = BM25Retriever.from_defaults(
            nodes=nodes, similarity_top_k=self.top_k
        )
        nodes = retriever.retrieve(query)
        return get_nodes_text(nodes)


class InteractiveXPathRetriever(BaseHtmlRetriever):
    def __init__(self, driver: BaseDriver):
        self.driver = driver

    def retrieve(
        self, query: QueryBundle, html_chunks: List[str], viewport_only=True
    ) -> List[str]:
        html = merge_html_chunks(html_chunks)
        possible_interactions = self.driver.get_possible_interactions(
            in_viewport=viewport_only
        )
        html = self.get_html_with_xpath(html, possible_interactions)
        return [html]

    def _generate_xpath(self, element, path=""):  # used to generate dict nodes
        """Recursive function to generate the xpath of an element"""
        if element.parent is None:
            return path
        else:
            siblings = [
                sib for sib in element.parent.children if sib.name == element.name
            ]
            tag = element.name
            if tag in ["svg", "path", "circle", "g"]:
                tag = f"*[local-name() = '{tag}']"
            if len(siblings) > 1:
                count = siblings.index(element) + 1
                if count == 1:
                    path = f"/{tag}{path}"
                else:
                    path = f"/{tag}[{count}]{path}"
            else:
                path = f"/{tag}{path}"
            return self._generate_xpath(element.parent, path)

    def get_html_with_xpath(
        self,
        html_content,
        filter_by_possible_interactions: Optional[PossibleInteractionsByXpath],
        xpath_prefix="",
    ):
        soup = BeautifulSoup(html_content, "html.parser")
        for element in soup.find_all(True):
            xpath = xpath_prefix + self._generate_xpath(element)
            if (
                filter_by_possible_interactions is None
                or xpath in filter_by_possible_interactions
            ):
                element["xpath"] = xpath
        for iframe_tag in soup.find_all("iframe"):
            frame_xpath = self._generate_xpath(iframe_tag)
            try:
                self.driver.switch_frame(frame_xpath)
            except Exception:
                continue
            frame_soup_str = self.get_html_with_xpath(
                self.driver.get_html(),
                filter_by_possible_interactions,
                xpath_prefix + frame_xpath,
            )
            iframe_tag.replace_with(frame_soup_str)
            self.driver.switch_parent_frame()
        return str(soup)


class OpsmSplitRetriever(BaseHtmlRetriever):
    def __init__(
        self,
        driver: BaseDriver,
        top_k: int = 5,
        group_by: int = 10,
        rank_fields: List[str] = ["element", "placeholder", "text", "name"],
    ):
        self.driver = driver
        self.top_k = top_k
        self.group_by = group_by
        self.rank_fields = rank_fields

    def _generate_xpath(self, element, path=""):  # used to generate dict nodes
        """Recursive function to generate the xpath of an element"""
        if element.parent is None:
            return path
        else:
            siblings = [
                sib for sib in element.parent.children if sib.name == element.name
            ]
            if len(siblings) > 1:
                count = siblings.index(element) + 1
                path = f"/{element.name}[{count}]{path}"
            else:
                path = f"/{element.name}{path}"
            return self._generate_xpath(element.parent, path)

    def _add_xpath_attributes(self, html_content):
        """
        Add an 'xpath' attribute to each element in the HTML content with its computed XPath.
        """
        soup = BeautifulSoup(html_content, "lxml")
        for element in soup.find_all(True):
            xpath = self._generate_xpath(element)
            element["xpath"] = xpath
        return str(soup)

    def _create_nodes_dict(
        self, html, only_body=True, max_length=200
    ):  # used to generate dict nodes
        """Create a list of xpaths and a list of dict of attributes of all elements in the html"""
        soup = BeautifulSoup(html, "html.parser")
        if only_body:
            root = soup.body
        else:
            root = soup.html
        element_attributes_list = []
        stack = [root]  # stack to keep track of elements and their paths
        while stack:
            element = stack.pop()
            if element.name is not None:
                element_attrs = dict(element.attrs)
                direct_text_content = "".join(
                    [
                        str(content).strip()
                        for content in element.contents
                        if isinstance(content, NavigableString) and content.strip()
                    ]
                )
                if direct_text_content:
                    element_attrs["text"] = direct_text_content
                    element_attrs["element"] = element.name
                    for key in element_attrs:
                        if len(element_attrs[key]) > max_length:
                            element_attrs[key] = element_attrs[key][:max_length]
                    element_attributes_list.append(element_attrs)
                elif element_attrs != {}:
                    element_attrs["element"] = element.name
                    for key in element_attrs:
                        if len(element_attrs[key]) > max_length:
                            element_attrs[key] = element_attrs[key][:max_length]
                    element_attributes_list.append(element_attrs)
                for child in element.children:
                    if child.name is not None:
                        stack.append(child)
        return element_attributes_list

    def _chunk_dicts(self, dicts, chunk_size=10):
        def chunks(lst, n):
            for i in range(0, len(lst), n):
                yield lst[i : i + n]

        grouped_chunks = []
        for chunk in chunks(dicts, chunk_size):
            all_keys = set(key for d in chunk for key in d.keys())
            grouped = {key: [] for key in all_keys}
            for d in chunk:
                for key in all_keys:
                    grouped[key].append(d.get(key, ""))
            grouped_chunks.append(grouped)
        return grouped_chunks

    def _unchunk_dicts(self, grouped_chunks):
        flat_list = []
        for group in grouped_chunks:
            max_length = max(len(v) for v in group.values())
            for i in range(max_length):
                new_dict = {}
                for key, values in group.items():
                    if i < len(values):
                        if values[i] != "":
                            new_dict[key] = values[i]
                if new_dict:
                    flat_list.append(new_dict)
        return flat_list

    def _clean_attributes(self, attributes_list):  # used to generate dict nodes
        if self.rank_fields:
            self.rank_fields.append("xpath")
            attributes_list = [
                {k: v for k, v in d.items() if k in self.rank_fields}
                for d in attributes_list
            ]
        attributes_list = [
            d
            for d in attributes_list
            if (
                not (
                    (len(list(d.keys())) == 2)
                    and (("element" in list(d.keys())) and "xpath" in list(d.keys()))
                )
            )
            or d == {}
        ]
        return attributes_list

    def _get_results(self, query, html):  # used to generate and retrieve dict nodes
        """Return the top_k elements of the html that are the most relevant to the query as Node objects with xpath in their metadata"""
        attributes_list = self._create_nodes_dict(html)
        # cleaning the attributes_list
        attributes_list = self._clean_attributes(attributes_list)
        # retrieving the top_k results

        list_of_results = []
        attributes_list = self._chunk_dicts(attributes_list, self.group_by)
        l = len(attributes_list)
        # grouping the attributes_list in groups of 1000 to avoid memory errors
        list_of_grouped_results = []
        for j in range(0, l, 1000):
            nodes = []
            attr = attributes_list[j : j + 1000]
            for d in attr:
                xpath = d.pop("xpath")
                nodes.append(TextNode(text=str(d), metadata={"xpath": xpath}))
            retriever = BM25Retriever.from_defaults(
                nodes=nodes, similarity_top_k=self.top_k
            )
            results = retriever.retrieve(query)
            list_of_grouped_results += results
        nodes = []
        for grouped_results in list_of_grouped_results:
            xpaths = grouped_results.metadata["xpath"]
            ds = self._unchunk_dicts([ast.literal_eval(grouped_results.text)])
            assert len(xpaths) == len(ds)
            for xpath, d in zip(xpaths, ds):
                nodes.append(TextNode(text=str(d), metadata={"xpath": xpath}))
        l2 = len(nodes)
        for j in range(0, l2, 1000):
            retriever = BM25Retriever.from_defaults(
                nodes=nodes[j : j + 1000], similarity_top_k=self.top_k
            )
            results = retriever.retrieve(query)
            list_of_results += results
        list_of_results = sorted(list_of_results, key=lambda x: x.score, reverse=True)
        results = list_of_results[: self.top_k]
        results_dict = [ast.literal_eval(r.text) for r in results]
        for i in range(len(results_dict)):
            results_dict[i]["xpath"] = results[i].metadata["xpath"]
        scores = [r.score for r in results]
        return results_dict, scores

    def _match_element(self, attributes, element_specs):
        i = 0
        for spec in element_specs:
            if attributes.get("xpath") == spec["xpath"]:
                return i
            i += 1
        return None

    def _return_nodes_with_xpath(self, nodes, results_dict, score):
        returned_nodes = []
        for node in nodes:
            split_html = node.text
            soup = BeautifulSoup(split_html, "html.parser")
            for element in soup.descendants:
                if not isinstance(element, NavigableString):
                    indice = self._match_element(element.attrs, results_dict)
                    if indice is not None:
                        node.metadata["score"] = score[indice]
                        returned_nodes.append(node)
                        break
        return returned_nodes

    def retrieve(
        self, query: QueryBundle, html_nodes: List[str], viewport_only=True
    ) -> List[str]:
        html = self._add_xpath_attributes(merge_html_chunks(html_nodes))
        text_list = [html]
        documents = [Document(text=t) for t in text_list]
        splitter = LangchainNodeParser(
            lc_splitter=RecursiveCharacterTextSplitter.from_language(
                language="html",
            )
        )
        nodes = splitter.get_nodes_from_documents(documents)
        results_dict, score = self._get_results(query.query_str, html)
        for r in results_dict:
            if not self.driver.check_visibility(r["xpath"]):
                i = results_dict.index(r)
                results_dict.remove(r)
                score.pop(i)
        results_nodes = self._return_nodes_with_xpath(nodes, results_dict, score)
        results = [
            NodeWithScore(node=node, score=node.metadata["score"])
            for node in results_nodes
        ]
        return get_nodes_text(results)


class XPathedChunkRetriever(BaseHtmlRetriever):
    def retrieve(
        self, query: QueryBundle, html_chunks: List[str], viewport_only=True
    ) -> List[str]:
        pattern = re.compile(r'xpath="([^"]+)"')
        interactive_chunks = []
        for chunk in html_chunks:
            match = re.search(pattern, chunk)
            if match:
                interactive_chunks.append(chunk)
        return interactive_chunks


class FromXPathNodesExpansionRetriever(BaseHtmlRetriever):
    """
    Retriever with expansion of HTML context around interactive elements (`chunk_size` in characters),
    then semantic contraction up to `top_k` results (number of chunks).
    Expansion is symmetrical so every step will add previous and next sibling.
    If no more sibling is present, then context is extended to parent node.
    When interactive chunks intersect, they are merged together.
    """

    def __init__(
        self,
        chunk_size: int = 750,
    ):
        self.chunk_size = chunk_size

    def get_included_xpaths(self, element) -> List[str]:
        if isinstance(element, NavigableString):
            return []
        xpaths = [e["xpath"] for e in element.find_all(attrs={"xpath": True})]
        if "xpath" in element.attrs:
            xpaths.append(element["xpath"])
        return xpaths

    def get_expanded_chunks(self, html_chunks: List[str]) -> List[str]:
        html = merge_html_chunks(html_chunks)
        soup = BeautifulSoup(html, "html.parser")
        elements = soup.find_all(attrs={"xpath": True})
        chunks = []
        processed_xpaths = set()

        def include_html(sibling) -> str:
            sibling_xpaths = self.get_included_xpaths(sibling)
            if processed_xpaths.isdisjoint(sibling_xpaths):
                processed_xpaths.update(sibling_xpaths)
                return str(sibling)
            return ""

        # For each marked element
        for element in elements:
            xpath = element["xpath"]
            if xpath in processed_xpaths:
                continue

            chunk = str(element)
            processed_xpaths.add(xpath)

            # Expand to siblings, then parent until we reach the chunk size
            while len(chunk) < self.chunk_size and (
                element.parent or element.previous_sibling or element.next_sibling
            ):
                previous_sibling = element.previous_sibling
                next_sibling = element.next_sibling

                # Add siblings to the chunk, from the closest to the farthest ones
                while len(chunk) < self.chunk_size and (
                    previous_sibling or next_sibling
                ):
                    if previous_sibling:
                        chunk = include_html(previous_sibling) + chunk
                        previous_sibling = previous_sibling.previous_sibling
                    if next_sibling:
                        chunk = chunk + include_html(next_sibling)
                        next_sibling = next_sibling.next_sibling

                # Move to parent if no more siblings can be added
                if len(chunk) < self.chunk_size and element.parent:
                    element = element.parent
                    chunk = str(element)
                    parent_xpaths = set(self.get_included_xpaths(element))
                    processed_xpaths.update(parent_xpaths)

                    # Remove previous chunks that are now included in the parent
                    chunks = [c for c in chunks if c not in chunk]

            if chunk.strip():
                chunks.append(chunk)

        return chunks

    def retrieve(
        self, query: QueryBundle, html_chunks: List[str], viewport_only=True
    ) -> List[str]:
        results = self.get_expanded_chunks(html_chunks)
        return results


class SemanticRetriever(BaseHtmlRetriever):
    """
    Semantic retriever up to `top_k` results (number of chunks)
    """

    def __init__(
        self,
        embedding: Optional[BaseEmbedding],
        top_k: int = 5,
        xpathed_only=True,
    ):
        self.top_k = top_k
        self.xpathed_only = xpathed_only
        self.embedding = embedding

    def retrieve(
        self, query: QueryBundle, html_chunks: List[str], viewport_only=True
    ) -> List[str]:
        splitter = LangchainNodeParser(
            lc_splitter=RecursiveCharacterTextSplitter.from_language(
                language="html",
            )
        )
        nodes = splitter.get_nodes_from_documents(
            [Document(text=merge_html_chunks(html_chunks))]
        )

        if self.xpathed_only:
            nodes = filter_for_xpathed_nodes(nodes)

        index = VectorStoreIndex(nodes=nodes, embed_model=self.embedding)
        query_engine = index.as_retriever(similarity_top_k=self.top_k)

        retrieved_nodes = query_engine.retrieve(query)
        return get_nodes_text(retrieved_nodes)


class SyntaxicRetriever(BaseHtmlRetriever):
    """
    Syntaxic retriever up to `top_k` results (number of chunks)
    """

    def __init__(
        self,
        top_k: int = 5,
        xpathed_only=True,
    ):
        self.top_k = top_k
        self.xpathed_only = xpathed_only

    def retrieve(
        self, query: QueryBundle, html_chunks: List[str], viewport_only=True
    ) -> List[str]:
        documents = [Document(text=merge_html_chunks(html_chunks))]
        splitter = LangchainNodeParser(
            lc_splitter=RecursiveCharacterTextSplitter.from_language(
                language="html",
            )
        )
        nodes = splitter.get_nodes_from_documents(documents)
        if self.xpathed_only:
            nodes = filter_for_xpathed_nodes(nodes)

        retriever = BM25Retriever.from_defaults(
            nodes=nodes, similarity_top_k=self.top_k
        )
        results = retriever.retrieve(query)
        return get_nodes_text(results)


def filter_for_xpathed_nodes(nodes: List):
    pattern = re.compile(r'xpath="([^"]+)"')
    compatibles = []
    for node in nodes:
        match = re.search(pattern, node.text)
        if match:
            compatibles.append(node)
    return compatibles if len(compatibles) > 0 else nodes


def get_nodes_text(nodes: List[NodeWithScore]) -> List[str]:
    return [n.text for n in nodes]


def merge_html_chunks(html_chunks: List[str], separator="\n") -> str:
    return separator.join(html_chunks)



from abc import ABC, abstractmethod
import pandas as pd
from typing import Dict
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.figure import Figure
from lavague.core.retrievers import BaseHtmlRetriever
from lavague.core.navigation import NavigationEngine
from lavague.drivers.selenium import SeleniumDriver
from tqdm import tqdm
from datetime import datetime
import yaml
from llama_index.core import QueryBundle
import traceback
import base64
import ast
from bs4 import BeautifulSoup


class Evaluator(ABC):
    @abstractmethod
    def evaluate(self) -> pd.DataFrame:
        pass

    def compare(
        self,
        results: Dict[str, pd.DataFrame],
        metrics: list,
    ) -> Figure:
        fig, axes = plt.subplots(1, len(metrics), figsize=(5 * len(metrics), 5))

        df = pd.DataFrame()
        for metric in metrics:
            if metric in metrics:
                df[metric] = [dfr[metric].mean() for dfr in results.values()]
        df["name"] = list(results.keys())

        count = 0
        for metric in metrics:
            if metric in metrics:
                plot = sns.barplot(data=df, x="name", y=metric, ax=axes[count])
                count += 1
                if metric == "time":
                    plot.set(xlabel="", ylabel="time (secs)")
                else:
                    plot.set(xlabel="")
        return fig


def parse_action(action: str) -> dict:
    action = yaml.safe_load(action)
    try:
        action = action[0]["actions"][0]["action"]
        _ = action["args"]["xpath"]
        _ = action["name"]
        return action
    except:
        return None


def parse_yaml(action):
    try:
        return yaml.safe_load(action)[0]["actions"][0]["action"]
    except:
        return None


def parse_viewport_size(vsize):
    vsize = ast.literal_eval(vsize)
    if type(vsize["width"]) is dict:
        return {
            "width": vsize["width"]["value"],
            "height": vsize["height"]["value"],
        }  # sometime viewport size is logged like this. idkw
    return vsize


def validate_action(action):
    try:
        _ = action["args"]["xpath"]
        _ = action["name"]
        return action["name"] != "fail"
    except:
        return False


def remove_img(html):
    soup = BeautifulSoup(html, "html.parser")
    for tag in soup.find_all("img"):
        tag.extract()
    return soup.decode()


FAIL_ACTION = {"args": {"xpath": "(string)"}, "name": "fail"}


class RetrieverEvaluator(Evaluator):
    def evaluate(
        self,
        retriever: BaseHtmlRetriever,
        dataset: pd.DataFrame,
        driver: SeleniumDriver = None,  # Optional, the driver passed to the retriever
        retriever_name: str = "",
    ) -> pd.DataFrame:
        result_filename = (
            (retriever_name if retriever_name else type(retriever).__name__)
            + "_evaluation_"
            + datetime.now().strftime("%Y-%m-%d_%H-%M")
            + ".csv"
        )
        results = dataset.loc[dataset["is_verified"]].copy()
        results.insert(len(results.columns), "recall", None)
        results.insert(len(results.columns), "output_size", None)
        results.insert(len(results.columns), "time", None)
        results["dataset_index"] = results.index
        results.to_csv(result_filename, index=False)

        try:
            for i, row in tqdm(results.iterrows()):
                driver.__init__()  # reinit the driver
                action = parse_action(row["action"])
                if driver:  # artificially get the page if the retriever needs a driver
                    html_bs64 = base64.b64encode(
                        remove_img(row["preaction_html_bundle"]).encode()
                    ).decode()
                    driver.get("data:text/html;base64," + html_bs64)
                    viewport_size = parse_viewport_size(row["viewport_size"])
                    driver.resize_driver(
                        width=viewport_size["width"], height=viewport_size["height"]
                    )
                instruction = row["instruction"]
                try:
                    t_begin = datetime.now()
                    nodes = retriever.retrieve(
                        QueryBundle(query_str=instruction), [driver.get_html()]
                    )
                    t_end = datetime.now()
                except:
                    print("ERROR: ", i)
                    traceback.print_exc()
                    nodes = []
                nodes = "\n".join(nodes)
                results.at[i, "recall"] = 1 if action["args"]["xpath"] in nodes else 0
                results.at[i, "output_size"] = len(nodes)
                results.at[i, "time"] = pd.Timedelta(t_end - t_begin).total_seconds()
            print("Evaluation terminated succesfully.")
        except:
            traceback.print_exc()
            print(f"Evaluation stopped at row {i} because an exception was caught.")
        finally:
            results.to_csv(result_filename, index=False)
            print(f"Results are saved to {result_filename}")
            return results

    def compare(
        self,
        results: Dict[str, pd.DataFrame],
        metrics: list = ["recall", "output_size", "time"],
    ) -> Figure:
        return super().compare(results, metrics)


class NavigationEngineEvaluator(Evaluator):
    def evaluate(
        self,
        navigation_engine: NavigationEngine,
        dataset: pd.DataFrame,
        navigation_engine_name="",
    ) -> pd.DataFrame:
        result_filename = (
            (
                navigation_engine_name
                if navigation_engine_name
                else type(navigation_engine).__name__
            )
            + "_evaluation_"
            + datetime.now().strftime("%Y-%m-%d_%H-%M")
            + ".csv"
        )
        results = dataset.loc[dataset["is_verified"]].copy()
        results.insert(len(results.columns), "recall", None)
        results.insert(len(results.columns), "correct_action", None)
        results.insert(len(results.columns), "correct_xpath", None)
        results.insert(len(results.columns), "time", None)
        results["dataset_index"] = results.index
        results.to_csv(result_filename, index=False)

        try:
            for i, row in tqdm(results.iterrows()):
                action = parse_action(row["action"])
                html_bs64 = base64.b64encode(
                    remove_img(row["preaction_html_bundle"]).encode()
                ).decode()
                navigation_engine.driver.get("data:text/html;base64," + html_bs64)
                viewport_size = parse_viewport_size(row["viewport_size"])
                navigation_engine.driver.resize_driver(
                    width=viewport_size["width"], height=viewport_size["height"]
                )
                instruction = row["instruction"]
                try:
                    t_begin = datetime.now()
                    test_action = navigation_engine.execute_instruction(
                        instruction
                    ).code
                    test_action = parse_yaml(test_action)
                    if not validate_action(test_action):
                        test_action = FAIL_ACTION
                    t_end = datetime.now()
                except:
                    print("ERROR: ", i)
                    traceback.print_exc()
                    test_action = FAIL_ACTION
                results.at[i, "correct_action"] = action["name"] == test_action["name"]
                results.at[i, "correct_xpath"] = (
                    action["args"]["xpath"] == test_action["args"]["xpath"]
                )
                results.at[i, "recall"] = (
                    results.at[i, "correct_action"] and results.at[i, "correct_xpath"]
                )
                results.at[i, "time"] = pd.Timedelta(t_end - t_begin).total_seconds()
            print("Evaluation terminated succesfully.")
        except:
            traceback.print_exc()
            print(f"Evaluation stopped at row {i} because an exception was caught.")
        finally:
            results.to_csv(result_filename, index=False)
            print(f"Results are saved to {result_filename}")
            return results

    def compare(
        self,
        results: Dict[str, pd.DataFrame],
        metrics: list = ["recall", "correct_action", "correct_xpath", "time"],
    ) -> Figure:
        return super().compare(results, metrics)



class NavigationException(Exception):
    pass


class CannotBackException(NavigationException):
    def __init__(self, message="History root reached, cannot go back"):
        super().__init__(message)


class RetrievalException(NavigationException):
    pass


class NoElementException(RetrievalException):
    def __init__(self, message="No element found"):
        super().__init__(message)


class AmbiguousException(RetrievalException):
    def __init__(self, message="Multiple elements could match"):
        super().__init__(message)


class HallucinatedException(RetrievalException):
    def __init__(self, xpath: str, message: str = None):
        super().__init__(message or f"Element was hallucinated: {xpath}")



from typing import Callable, List
import inspect
import re
import ast

DEFAULT_ENGINES: List[str] = [
    "Navigation Controls",
    "Python Engine",
    "Navigation Engine",
    "COMPLETE",
]


class VariableVisitor(ast.NodeVisitor):
    """Helper class to visit AST nodes and extract variables assigned in the code."""

    def __init__(self):
        super().__init__()
        self.output = []

    def visit_Assign(self, node):
        # For each assignment, print the targets (variables)
        for target in node.targets:
            if isinstance(target, ast.Name):  # Ensure it's a variable assignment
                self.output.append(target.id)


def quote_numeric_yaml_values(yaml_string):
    """Wrap numeric values in quotes in a YAML string.

    This is helpful when the YAML might contain numeric values that are not wrapped in quotes which can cause issues when parsing the YAML.
    In Navigation Engine, we expect values to be always strings, so this avoids issues when the YAML does not wrap numeric values in quotes, such as outputting "value: 01" instead of "value: '01'".

    """

    def replace_value(match):
        full_match, value = match.groups()
        try:
            # Try to convert to float to catch both integers and decimal numbers
            float(value)
            # If successful, it's a number, so wrap it in quotes
            return f'value: "{value}"'
        except ValueError:
            # If it's not a number, return the original match
            return full_match

    # Regex to match 'value:' followed by a space and then any numeric
    pattern = r"(value: ([\d.]+))"

    # Replace values that are numeric
    modified_yaml = re.sub(pattern, replace_value, yaml_string)

    return modified_yaml


def return_assigned_variables(code_snippet):
    """Returns the variables assigned in a code snippet."""
    parsed = ast.parse(code_snippet)

    # Create a VariableVisitor object
    visitor = VariableVisitor()

    # Visit (i.e., traverse) the AST nodes
    visitor.visit(parsed)

    return visitor.output


def keep_assignments(code_snippet):
    # Regex to match variable assignments. This pattern assumes variable names are valid Python identifiers
    # and captures typical assignment statements, excluding those that might appear in comments or strings.
    pattern = r"^\s*[a-zA-Z_][a-zA-Z0-9_]*\s*=\s*.+"

    # Filter and keep only lines with variable assignments
    assignments = [line for line in code_snippet.split("\n") if re.match(pattern, line)]

    # Join the filtered lines back into a string
    return "\n".join(assignments)


def extract_code_from_funct(funct: Callable) -> List[str]:
    """Extract code lines from a function while removing the first line (function definition) and the last line (return) and correcting indentation"""
    source_code = inspect.getsource(funct)
    source_code_lines = source_code.splitlines()[1:]  # remove the first line
    nident = len(source_code_lines[0]) - len(
        source_code_lines[0].lstrip()
    )  # count nb char in indentation
    return [
        line[nident:] for line in source_code_lines[:-1]
    ]  # every line except the return


def extract_imports_from_lines(lines: List[str]) -> str:
    """Only keep import lines from python code lines and join them"""
    return "\n".join(
        [line for line in lines if line.startswith("from") or line.startswith("import")]
    )


import re


def extract_world_model_instruction(text):
    # Use a regular expression to find the content after "Instruction:"
    instruction_patterns = [
        r"Instruction:\s*((?:- .*\n?)+)",  # For multi-line hyphenated instructions
        r"### Instruction:\s*((?:- .*\n?)+)",  # For multi-line hyphenated instructions with ### prefix
        r"Instruction:\s*((?:\d+\.\s.*\n?)+)",  # For multi-line numbered instructions
        r"### Instruction:\s*((?:\d+\.\s.*\n?)+)",  # For multi-line numbered instructions with ### prefix
        r"Instruction:\s*```(.*?)```",  # For block of text within triple backticks
        r"### Instruction:\s*```(.*?)```",  # For block of text within triple backticks with ### prefix
        r"Instruction:\s*(.*)",  # For single-line instructions
        r"### Instruction:\s*(.*)",  # For single-line instructions with ### prefix
    ]

    longest_instruction = ""

    for pattern in instruction_patterns:
        matches = re.findall(pattern, text, re.MULTILINE | re.DOTALL)
        for instruction_text in matches:
            instruction_text = instruction_text
            # Check if the instruction is multi-line or single-line
            if "\n" in instruction_text:
                # Remove newlines and extra spaces for multi-line instructions
                instruction_str = " ".join(
                    line for line in instruction_text.split("\n")
                )
            else:
                instruction_str = instruction_text
            # Update longest_instruction if the current one is longer
            if len(instruction_str) > len(longest_instruction):
                longest_instruction = instruction_str

    if longest_instruction:
        return longest_instruction

    raise ValueError("No instruction found in the text.")


def replace_hyphens(text: str, replacement_char="•"):
    formatted_text = text.replace("- ", f"{replacement_char} ")
    return formatted_text


def extract_before_next_engine(text: str) -> str:
    # Define the patterns for "Next engine:" and similar patterns
    next_engine_patterns = [r"Next engine:\s*", r"### Next Engine:\s*"]

    # Split the text using the "Next engine:" patterns
    for pattern in next_engine_patterns:
        split_text = re.split(pattern, text, maxsplit=1)
        if len(split_text) > 1:
            result = split_text[0].strip()
            break
    else:
        result = text.strip()

    thoughts_pattern = r"^Thoughts:\s*"
    result = re.sub(thoughts_pattern, "", result).strip()

    return result


def extract_next_engine(text: str, next_engines: List[str] = DEFAULT_ENGINES) -> str:
    # Use a regular expression to find the content after "Next engine:"

    next_engine_patterns = [r"Next engine:\s*(.*)", r"### Next Engine:\s*(.*)"]

    for pattern in next_engine_patterns:
        next_engine_match = re.search(pattern, text)
        if next_engine_match:
            extracted_text = next_engine_match.group(1).strip()
            # To avoid returning a non-existent engine

            for engine in next_engines:
                if engine.lower() in extracted_text.lower():
                    return engine

    raise ValueError(f"No next engine found in the text: {text}")


def extract_and_eval(string, extract_list=True):
    # Regular expression to match a list inside a string
    if extract_list:
        regex = r"\[.*?\]"
    else:
        regex = r"\{.*?\}"
    match = re.search(regex, string, re.DOTALL)
    if match:
        list_string = match.group(0)
        try:
            result = ast.literal_eval(list_string)
            return result
        except (SyntaxError, ValueError):
            raise ValueError(
                f"The extracted string is not a valid Python literal: {list_string}"
            )
    else:
        raise ValueError(f"No list can be extracted from: {match}")


def clean_html(
    html_to_clean: str,
    tags_to_remove: List[str] = ["style", "svg", "script"],
    attributes_to_keep: List[str] = ["id", "href"],
) -> str:
    """
    Clean HTML content by removing specified tags and attributes while keeping specified attributes.

    Args:
        html_to_clean (str): The HTML content to clean.
        tags_to_remove (List[str]): List of tags to remove from the HTML content. Default is ['style', 'svg', 'script'].
        attributes_to_keep (List[str]): List of attributes to keep in the HTML tags. Default is ['id', 'href'].

    Returns:
        str: The cleaned HTML content.

    Example:
    >>> from clean_html_for_llm import clean_html
    >>> cleaned_html = clean_html('<div id="main" style="color:red">Hello <script>alert("World")</script></div>', tags_to_remove=['script'], attributes_to_keep=['id'])
    """
    for tag in tags_to_remove:
        html_to_clean = re.sub(
            rf"<{tag}[^>]*>.*?</{tag}>", "", html_to_clean, flags=re.DOTALL
        )

    attributes_to_keep = "|".join(attributes_to_keep)
    pattern = rf'\b(?!({attributes_to_keep})\b)\w+(?:-\w+)?\s*=\s*["\'][^"\']*["\']'
    cleaned_html = re.sub(pattern, "", html_to_clean)
    return cleaned_html



import logging
import os
from typing import Any, Dict
from pandas import DataFrame
import uuid
import msgpack
import numpy as np
import requests

from .version_checker import get_installed_version

TELEMETRY_VAR = os.getenv("LAVAGUE_TELEMETRY")
UNIQUE_ID = os.getenv("LAVAGUE_UNIQUE_USER_ID")
USER_ID = str(uuid.uuid4())

if UNIQUE_ID is not None:
    UNIQUE_ID = UNIQUE_ID[:256]

logging_print = logging.getLogger(__name__)
logging_print.setLevel(logging.INFO)
format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(format)
logging_print.addHandler(ch)
logging_print.propagate = False


def send_telemetry(logger_telemetry: DataFrame, origin: str, test: bool = False):
    try:
        if TELEMETRY_VAR is None:
            logger_telemetry = logger_telemetry.drop(
                "screenshots", axis=1, errors="ignore"
            )
            logger_telemetry = logger_telemetry.drop(
                "screenshots_path", axis=1, errors="ignore"
            )
            logger_telemetry = logger_telemetry.drop("html", axis=1, errors="ignore")
            logger_telemetry = logger_telemetry.replace({np.nan: None})

            logger_telemetry["origin"] = origin

            for index, row in logger_telemetry.iterrows():
                logger_telemetry.at[index, "unique_user_id"] = UNIQUE_ID
                logger_telemetry.at[index, "user_id"] = USER_ID
                logger_telemetry.at[index, "version"] = get_installed_version(
                    "lavague-core"
                )

                if "engine_log" in row:
                    t = row["engine_log"]
                    if t is not None:
                        if isinstance(t, list):
                            for t_obj in t:
                                if "vision_data" in t_obj:
                                    vision = t_obj["vision_data"]
                                    if vision is not None:
                                        for i in range(len(vision)):
                                            if "screenshot" in vision[i]:
                                                del vision[i]["screenshot"]
                            logger_telemetry.at[index, "engine_log"] = t
                        else:
                            if "vision_data" in t:
                                vision = t["vision_data"]
                                for i in range(len(vision)):
                                    if "screenshot" in vision[i]:
                                        del vision[i]["screenshot"]
                                logger_telemetry.at[index, "engine_log"] = t

            dic = logger_telemetry.to_dict("records")
            pack = msgpack.packb(dic)

            r = requests.post(
                "https://telemetrylavague.mithrilsecurity.io/telemetry_new", data=pack
            )
            if r.status_code != 200:
                raise ValueError(r.content)
            else:
                logging_print.debug("Telemetry sent successfully")
        elif TELEMETRY_VAR == "NONE":
            pass
    except Exception as e:
        if not test:
            logging_print.warning(f"Telemetry failed with {e}")
        else:
            raise ValueError("Telemetry failed with ", e)



from __future__ import annotations
from typing import Dict
from llama_index.core import PromptTemplate
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.base.embeddings.base import BaseEmbedding
from lavague.core.extractors import BaseExtractor, DynamicExtractor
from lavague.core.retrievers import BaseHtmlRetriever, get_default_retriever
from lavague.core.base_driver import BaseDriver
from lavague.core.context import Context, get_default_context
from lavague.core.logger import AgentLogger
from lavague.core.base_engine import BaseEngine, ActionResult
from lavague.core.navigation import NAVIGATION_ENGINE_PROMPT_TEMPLATE
from lavague.core.navigation import NavigationControl, NavigationEngine
from lavague.core.python_engine import PythonEngine
from lavague.core.utilities.model_utils import get_model_name


class ActionEngine:
    """
    ActionEngine is a wrapper that instantiate every other engines (Navigation Engine, Python Engine, Navigation Control)

    Args:
        driver (`BaseDriver`):
            The Web driver used to interact with the headless browser
        python_engine (`BaseActionEngine`)
            Python Engine for generating code that doesn't interact with an html page.
        navigation_control (`BaseActionEngine`)
            Navigation Control
        llm (`BaseLLM`)
            llama-index LLM that will generate the action
        embedding (`BaseEmbedding`)
            llama-index Embedding model
        retriever (`BaseHtmlRetriever`)
            Specify which algorithm will be used for RAG
        prompt_template (`PromptTemplate`)
            Squelette of the final prompt
        extractor (`BaseExtractor`)
            Specify how to extract the final code from the llm answer
        time_between_actions (`float`)
            Time between each action
        logger: (`AgentLogger`)
            Logger to log the actions taken by the agent
    """

    def __init__(
        self,
        driver: BaseDriver,
        navigation_engine: BaseEngine = None,
        python_engine: BaseEngine = None,
        navigation_control: BaseEngine = None,
        llm: BaseLLM = None,
        embedding: BaseEmbedding = None,
        retriever: BaseHtmlRetriever = None,
        prompt_template: PromptTemplate = NAVIGATION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = DynamicExtractor(),
        time_between_actions: float = 1.5,
        n_attempts: int = 5,
        logger: AgentLogger = None,
    ):
        if llm is None:
            llm = get_default_context().llm

        if embedding is None:
            embedding = get_default_context().embedding

        self.driver = driver

        if retriever is None:
            retriever = get_default_retriever(driver, embedding=embedding)

        if navigation_engine is None:
            navigation_engine = NavigationEngine(
                driver=driver,
                llm=llm,
                retriever=retriever,
                prompt_template=prompt_template,
                extractor=extractor,
                time_between_actions=time_between_actions,
                n_attempts=n_attempts,
                logger=logger,
                embedding=embedding,
            )
        if python_engine is None:
            python_engine = PythonEngine(driver, llm, embedding)
        if navigation_control is None:
            navigation_control = NavigationControl(
                driver,
                time_between_actions=time_between_actions,
                navigation_engine=navigation_engine,
            )
        self.navigation_engine = navigation_engine
        self.python_engine = python_engine
        self.navigation_control = navigation_control
        self.engines: Dict[str, BaseEngine] = {
            "Navigation Engine": self.navigation_engine,
            "Python Engine": self.python_engine,
            "Navigation Controls": self.navigation_control,
        }
        self.ret = None
        self.highlight = None
        self.curr_step = 0
        self.curr_instruction = ""
        self.world_model_output = ""
        self.screenshot_ratio = 1

    @classmethod
    def from_context(
        cls,
        context: Context,
        driver: BaseDriver,
        navigation_engine: BaseEngine = None,
        python_engine: BaseEngine = None,
        navigation_control: BaseEngine = None,
        retriever: BaseHtmlRetriever = None,
        prompt_template: PromptTemplate = NAVIGATION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = DynamicExtractor(),
        time_between_actions: float = 1.5,
        n_attempts: int = 5,
        logger: AgentLogger = None,
    ) -> ActionEngine:
        """
        Create an ActionEngine from a context
        """
        return cls(
            driver,
            navigation_engine,
            python_engine,
            navigation_control,
            context.llm,
            context.embedding,
            retriever,
            prompt_template,
            extractor,
            time_between_actions,
            n_attempts,
            logger,
        )

    def set_gradio_mode_all(
        self,
        gradio_mode: bool,
        objective,
        url_input,
        image_display,
        history,
    ):
        self.navigation_engine.set_gradio_mode(
            gradio_mode,
            objective,
            url_input,
            image_display,
            history,
        )
        self.python_engine.set_gradio_mode(
            gradio_mode,
            objective,
            url_input,
            image_display,
            history,
        )
        self.navigation_control.set_gradio_mode(
            gradio_mode,
            objective,
            url_input,
            image_display,
            history,
        )

    def set_display_all(self, display: bool):
        self.navigation_engine.set_display(display)
        self.python_engine.set_display(display)
        self.navigation_control.set_display(display)

    def set_logger_all(self, logger: AgentLogger):
        self.navigation_engine.set_logger(logger)
        self.python_engine.set_logger(logger)
        self.navigation_control.set_logger(logger)

    def dispatch_instruction_gradio(self, next_engine_name: str, instruction: str):
        """
        Dispatch the instruction to the appropriate ActionEngine

        Args:
            next_engine_name (`str`): The name of the engine to call
            instruction (`str`): The instruction to perform

        Return:
            `bool`: True if the code was executed without error
            `Any`: The output of the code
        """

        next_engine = self.engines[next_engine_name]

        if next_engine_name == "Navigation Engine":
            yield from next_engine.execute_instruction_gradio(instruction, self)
        else:
            ret = next_engine.execute_instruction(instruction)
            self.ret = ret
            yield (
                self.navigation_engine.objective,
                self.navigation_engine.url_input,
                self.navigation_engine.image_display,
                self.navigation_engine.history,
                ret.output,
            )

    def dispatch_instruction(
        self, next_engine_name: str, instruction: str
    ) -> ActionResult:
        """
        Dispatch the instruction to the appropriate ActionEngine

        Args:
            next_engine_name (`str`): The name of the engine to call
            instruction (`str`): The instruction to perform

        Return:
            `bool`: True if the code was executed without error
            `Any`: The output of the code
        """

        next_engine = self.engines[next_engine_name]
        return next_engine.execute_instruction(instruction)

    def get_llm_name(self):
        return get_model_name(self.python_engine.llm)

    def get_embedding_name(self):
        return get_model_name(self.python_engine.embedding)



from abc import ABC, abstractmethod
import re
from jsonschema import validate, ValidationError
import yaml
import json
from typing import Any, Dict


class ExtractionError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    def __str__(self) -> str:
        return f"Error extracting the object: {self.args[0]}"


class BaseExtractor(ABC):
    @abstractmethod
    def extract(self, text: str) -> str:
        pass

    @abstractmethod
    def extract_as_object(self, text: str) -> Any:
        pass


class YamlFromMarkdownExtractor(BaseExtractor):
    """
    Extractor for the prompts that end with (or similar to) the following:

    --------------------------------------------
    Completion:
    --------------------------------------------
    """

    def extract(self, markdown_text: str) -> str:
        yml_str = markdown_text.strip()
        # Pattern to match the first ```yaml ``` code block
        pattern = r"```(?:yaml|yml|\n)(.*?)```"

        # Using re.DOTALL to make '.' match also newlines
        match = re.search(pattern, markdown_text, re.DOTALL)
        if match:
            # Return the first matched group, which is the code inside the ```python ```
            yml_str = match.group(1).strip()
        try:
            yaml.safe_load(yml_str)
            return yml_str
        except yaml.YAMLError:
            return None

    def extract_as_object(self, text: str):
        return yaml.safe_load(self.extract(text))


class JsonFromMarkdownExtractor(BaseExtractor):
    """
    Extractor for the prompts that end with (or similar to) the following:

    --------------------------------------------
    Completion:
    --------------------------------------------
    """

    def extract(self, markdown_text: str, shape_validator=None) -> str:
        # Pattern to match the first ```json ``` code block
        pattern = r"```json(.*?)```"

        # Using re.DOTALL to make '.' match also newlines
        match = re.search(pattern, markdown_text, re.DOTALL)

        if shape_validator:
            try:
                # checks if the json returned from the llm matchs the schema
                validate(
                    instance=json.loads(match.group(1).strip()), schema=shape_validator
                )
            except json.JSONDecodeError:
                raise ExtractionError("Invalid JSON format")
            except ValidationError:
                raise ExtractionError("JSON does not match schema")
        else:
            if match:
                # Return the first matched group, which is the code inside the ```python ```
                return match.group(1).strip()
            else:
                # Return None if no match is found
                return None

    def extract_as_object(self, text: str):
        return json.loads(self.extract(text))


class PythonFromMarkdownExtractor(BaseExtractor):
    """
    Extractor for the prompts that end with (or similar to) the following:

    --------------------------------------------
    Completion:
    --------------------------------------------
    """

    def extract(self, markdown_text: str) -> str:
        # Pattern to match the first ```python ``` code block
        pattern = r"```python(.*?)```"

        # Using re.DOTALL to make '.' match also newlines
        match = re.search(pattern, markdown_text, re.DOTALL)
        if match:
            # Return the first matched group, which is the code inside the ```python ```
            return match.group(1).strip()
        else:
            # Return None if no match is found
            return None

    def extract_as_object(self, text: str):
        return eval(self.extract(text))


class UntilEndOfMarkdownExtractor(BaseExtractor):
    """
    Extractor for the prompts that end with (or similar to) the following:

    --------------------------------------------
    Completion:
    ```python
    # Let's proceed step by step.
    --------------------------------------------
    """

    def extract(self, text: str) -> str:
        return text.split("```")[0]

    def extract_as_object(self, text: str) -> Any:
        return self.extract(text)


class DynamicExtractor(BaseExtractor):
    """
    Extractor for typed markdown blocks
    """

    def __init__(self):
        self.extractors: Dict[str, BaseExtractor] = {
            "json": JsonFromMarkdownExtractor(),
            "yaml": YamlFromMarkdownExtractor(),
            "yml": YamlFromMarkdownExtractor(),
            "python": PythonFromMarkdownExtractor(),
        }

    def get_type(self, text: str) -> str:
        types_pattern = "|".join(self.extractors.keys())
        pattern = rf"```({types_pattern}).*?```"
        match = re.search(pattern, text, re.DOTALL)
        if match:
            return match.group(1).strip()
        else:
            # Try to auto-detect first matching extractor
            for type, extractor in self.extractors.items():
                try:
                    value = extractor.extract(text)
                    if value:
                        return type
                except:
                    pass
            raise ValueError(f"No extractor pattern can be found from {text}")

    def extract(self, text: str) -> str:
        type = self.get_type(text)
        return self.extractors[type].extract(text)

    def extract_as_object(self, text: str) -> Any:
        type = self.get_type(text)
        return self.extractors[type].extract_as_object(text)



from PIL import Image
import os
from pathlib import Path
import re
from typing import Any, Callable, Optional, Mapping, Dict, Set, List, Union
from abc import ABC, abstractmethod
from lavague.core.utilities.format_utils import (
    extract_code_from_funct,
    extract_imports_from_lines,
)
from enum import Enum
import time
from datetime import datetime
import hashlib


class InteractionType(Enum):
    CLICK = "click"
    HOVER = "hover"
    SCROLL = "scroll"
    TYPE = "type"


PossibleInteractionsByXpath = Dict[str, Set[InteractionType]]

r_get_xpaths_from_html = r'xpath=["\'](.*?)["\']'


class BaseDriver(ABC):
    def __init__(self, url: Optional[str], init_function: Optional[Callable[[], Any]]):
        """Init the driver with the init funtion, and then go to the desired url"""
        self.init_function = (
            init_function if init_function is not None else self.default_init_code
        )
        self.driver = self.init_function()

        # Flag to check if the page has been previously scanned to avoid erasing screenshots from previous scan
        self.previously_scanned = False

        # extract import lines for later exec of generated code
        init_lines = extract_code_from_funct(self.init_function)
        self.import_lines = extract_imports_from_lines(init_lines)

        if url is not None:
            self.get(url)

    @abstractmethod
    def default_init_code(self) -> Any:
        """Init the driver, with the imports, since it will be pasted to the beginning of the output"""
        pass

    @abstractmethod
    def code_for_init(self) -> str:
        """Extract the code to past to the begining of the final script from the init code"""
        pass

    @abstractmethod
    def destroy(self) -> None:
        """Cleanly destroy the underlying driver"""
        pass

    @abstractmethod
    def get_driver(self) -> Any:
        """Return the expected variable name and the driver object"""
        pass

    @abstractmethod
    def resize_driver(driver, width, height):
        """
        Resize the driver to a targeted height and width.
        """

    @abstractmethod
    def get_url(self) -> Optional[str]:
        """Get the url of the current page"""
        pass

    @abstractmethod
    def get(self, url: str) -> None:
        """Navigate to the url"""
        pass

    @abstractmethod
    def code_for_get(self, url: str) -> str:
        """Return the code to navigate to the url"""
        pass

    @abstractmethod
    def back(self) -> None:
        """Navigate back"""
        pass

    @abstractmethod
    def maximize_window(self) -> None:
        pass

    @abstractmethod
    def code_for_back(self) -> None:
        """Return driver specific code for going back"""
        pass

    @abstractmethod
    def get_html(self, clean: bool = True) -> str:
        """
        Returns the HTML of the current page.
        If clean is True, We remove unnecessary tags and attributes from the HTML.
        Clean HTMLs are easier to process for the LLM.
        """
        pass

    def get_tabs(self) -> str:
        """Return description of the tabs opened with the current tab being focused.

        Example of output:
        Tabs opened:
        0 - Overview - OpenAI API
        1 - [CURRENT] Nos destinations Train - SNCF Connect
        """
        return "Tabs opened:\n 0 - [CURRENT] tab"

    def switch_tab(self, tab_id: int) -> None:
        """Switch to the tab with the given id"""
        pass

    def switch_frame(self, xpath) -> None:
        """
        switch to the frame pointed at by the xpath
        """
        raise NotImplementedError()

    def switch_default_frame(self) -> None:
        """
        Switch back to the default frame
        """
        raise NotImplementedError()

    def switch_parent_frame(self) -> None:
        """
        Switch back to the parent frame
        """
        raise NotImplementedError()

    def resolve_xpath(self, xpath):
        """
        Return the element for the corresponding xpath, the underlying driver may switch iframe if necessary
        """
        pass

    def save_screenshot(self, current_screenshot_folder: Path) -> str:
        """Save the screenshot data to a file and return the path. If the screenshot already exists, return the path. If not save it to the folder."""

        new_screenshot = self.get_screenshot_as_png()
        hasher = hashlib.md5()
        hasher.update(new_screenshot)
        new_hash = hasher.hexdigest()
        new_screenshot_name = f"{new_hash}.png"
        new_screenshot_full_path = current_screenshot_folder / new_screenshot_name

        # If the screenshot does not exist, save it
        if not new_screenshot_full_path.exists():
            with open(new_screenshot_full_path, "wb") as f:
                f.write(new_screenshot)
        return str(new_screenshot_full_path)

    def is_bottom_of_page(self) -> bool:
        return self.execute_script(
            "return (window.innerHeight + window.scrollY + 1) >= document.body.scrollHeight;"
        )

    def get_screenshots_whole_page(self) -> list[str]:
        """Take screenshots of the whole page"""
        screenshot_paths = []

        current_screenshot_folder = self.get_current_screenshot_folder()

        while True:
            # Saves a screenshot
            screenshot_path = self.save_screenshot(current_screenshot_folder)
            screenshot_paths.append(screenshot_path)
            self.execute_script("window.scrollBy(0, (window.innerHeight / 1.5));")
            self.wait_for_idle()

            if self.is_bottom_of_page():
                break

        self.previously_scanned = True
        return screenshot_paths

    @abstractmethod
    def get_possible_interactions(
        self, in_viewport=True, foreground_only=True
    ) -> PossibleInteractionsByXpath:
        """Get elements that can be interacted with as a dictionary mapped by xpath"""
        pass

    def check_visibility(self, xpath: str) -> bool:
        pass

    @abstractmethod
    def get_highlighted_element(self, generated_code: str):
        """Return the page elements that generated code interact with"""
        pass

    @abstractmethod
    def exec_code(
        self,
        code: str,
        globals: dict[str, Any] = None,
        locals: Mapping[str, object] = None,
    ):
        """Exec generated code"""
        pass

    @abstractmethod
    def execute_script(self, js_code: str) -> Any:
        """Exec js script in DOM"""
        pass

    @abstractmethod
    def scroll_up(self):
        pass

    @abstractmethod
    def scroll_down(self):
        pass

    @abstractmethod
    def code_for_execute_script(self, js_code: str):
        """return driver specific code to execute js script in DOM"""
        pass

    @abstractmethod
    def get_capability(self) -> str:
        """Prompt to explain the llm which style of code he should output and which variables and imports he should expect"""
        pass

    def get_obs(self) -> dict:
        """Get the current observation of the driver"""
        current_screenshot_folder = self.get_current_screenshot_folder()

        if not self.previously_scanned:
            # If the last operation was not to scan the whole page, we clear the screenshot folder
            try:
                if os.path.isdir(current_screenshot_folder):
                    for filename in os.listdir(current_screenshot_folder):
                        file_path = os.path.join(current_screenshot_folder, filename)
                        try:
                            # Check if it's a file and then delete it
                            if os.path.isfile(file_path) or os.path.islink(file_path):
                                os.remove(file_path)
                        except Exception as e:
                            print(f"Failed to delete {file_path}. Reason: {e}")

            except Exception as e:
                raise Exception(f"Error while clearing screenshot folder: {e}")
        else:
            # If the last operation was to scan the whole page, we reset the flag
            self.previously_scanned = False

        # We take a screenshot and computes its hash to see if it already exists
        self.save_screenshot(current_screenshot_folder)

        url = self.get_url()
        html = self.get_html()
        obs = {
            "html": html,
            "screenshots_path": str(current_screenshot_folder),
            "url": url,
            "date": datetime.now().isoformat(),
            "tab_info": self.get_tabs(),
        }

        return obs

    def wait(self, duration):
        import time

        time.sleep(duration)

    def wait_for_idle(self):
        pass

    def get_current_screenshot_folder(self) -> Path:
        url = self.get_url()

        if url is None:
            url = "blank"

        screenshots_path = Path("./screenshots")
        screenshots_path.mkdir(exist_ok=True)

        current_url = url.replace("://", "_").replace("/", "_")
        hasher = hashlib.md5()
        hasher.update(current_url.encode("utf-8"))

        current_screenshot_folder = screenshots_path / hasher.hexdigest()
        current_screenshot_folder.mkdir(exist_ok=True)
        return current_screenshot_folder

    @abstractmethod
    def get_screenshot_as_png(self) -> bytes:
        pass

    def get_nodes(self, xpaths: List[str]) -> List["DOMNode"]:
        raise NotImplementedError("get_nodes not implemented")

    def get_nodes_from_html(self, html: str) -> List["DOMNode"]:
        return self.get_nodes(re.findall(r_get_xpaths_from_html, html))

    def highlight_node_from_xpath(self, xpath: str, color: str = "red") -> Callable:
        return self.highlight_nodes([xpath], color)

    def highlight_nodes(self, xpaths: List[str], color: str = "red") -> Callable:
        nodes = [n.highlight(color) for n in self.get_nodes(xpaths)]
        return self._add_highlighted_destructors(lambda: [n.clear() for n in nodes])

    def highlight_nodes_from_html(self, html: str, color: str = "blue") -> Callable:
        return self.highlight_nodes(re.findall(r_get_xpaths_from_html, html), color)

    def remove_highlight(self):
        if hasattr(self, "_highlight_destructors"):
            for destructor in self._highlight_destructors:
                destructor()
            delattr(self, "_highlight_destructors")

    def _add_highlighted_destructors(
        self, destructors: Union[List[Callable], Callable]
    ):
        if not hasattr(self, "_highlight_destructors"):
            self._highlight_destructors = []
        if isinstance(destructors, Callable):
            self._highlight_destructors.append(destructors)
        else:
            self._highlight_destructors.extend(destructors)
        return destructors

    def highlight_interactive_nodes(
        self,
        *with_interactions: tuple[InteractionType],
        color: str = "red",
        in_viewport=True,
        foreground_only=True,
    ):
        if with_interactions is None or len(with_interactions) == 0:
            return self.highlight_nodes(
                list(
                    self.get_possible_interactions(
                        in_viewport=in_viewport, foreground_only=foreground_only
                    ).keys()
                ),
                color,
            )

        return self.highlight_nodes(
            [
                xpath
                for xpath, interactions in self.get_possible_interactions(
                    in_viewport=in_viewport, foreground_only=foreground_only
                ).items()
                if set(interactions) & set(with_interactions)
            ],
            color,
        )


class DOMNode(ABC):
    @abstractmethod
    def highlight(self, color: str = "red"):
        pass

    @abstractmethod
    def clear(self):
        return self

    @abstractmethod
    def take_screenshot(self) -> Image:
        pass

    @abstractmethod
    def get_html(self) -> str:
        pass

    def __str__(self) -> str:
        return self.get_html()


def js_wrap_function_call(fn: str):
    return "(function(){" + fn + "})()"


JS_SETUP_GET_EVENTS = """
(function() {
  if (window && !window.getEventListeners) {
    const targetProto = EventTarget.prototype;
    targetProto._addEventListener = Element.prototype.addEventListener;
    targetProto.addEventListener = function(a,b,c) {
        this._addEventListener(a,b,c);
        if(!this.eventListenerList) this.eventListenerList = {};
        if(!this.eventListenerList[a]) this.eventListenerList[a] = [];
        this.eventListenerList[a].push(b);
    };
    targetProto._removeEventListener = Element.prototype.removeEventListener;
    targetProto.removeEventListener = function(a, b, c) {
        this._removeEventListener(a, b, c);
        if(this.eventListenerList && this.eventListenerList[a]) {
        const index = this.eventListenerList[a].indexOf(b);
        if (index > -1) {
            this.eventListenerList[a].splice(index, 1);
            if(!this.eventListenerList[a].length) {
            delete this.eventListenerList[a];
            }
        }
        }
    };
    window.getEventListeners = function(e) {
      return (e && e.eventListenerList) || [];
    }
  }
})();"""

JS_GET_INTERACTIVES = """
return (function() {
    function getInteractions(e) {
        const tag = e.tagName.toLowerCase();
        if (!e.checkVisibility() || e.hasAttribute('disabled') || e.hasAttribute('readonly') || e.getAttribute('aria-hidden') === 'true'
          || e.getAttribute('aria-disabled') === 'true' || (tag === 'input' && e.getAttribute('type') === 'hidden')) {
            return [];
        }
        const rect = e.getBoundingClientRect();
        if (rect.width + rect.height < 5) {
            return [];
        }
        const style = getComputedStyle(e) || {};
        if (style.display === 'none' || style.visibility === 'hidden' || style.opacity === '0') {
            return [];
        }
        const events = window && typeof window.getEventListeners === 'function' ? window.getEventListeners(e) : [];
        const role = e.getAttribute('role');
        const clickableInputs = ['submit', 'checkbox', 'radio', 'color', 'file', 'image', 'reset'];
        function hasEvent(n) {
            return events[n]?.length || e.hasAttribute('on' + n);
        }
        const evts = [];
        if (hasEvent('keydown') || hasEvent('keyup') || hasEvent('keypress') || hasEvent('keydown') || hasEvent('input') || e.isContentEditable
          || (
            (tag === 'input' || tag === 'textarea' || role === 'searchbox' || role === 'input')
            ) && !clickableInputs.includes(e.getAttribute('type'))
          ) {
            evts.push('TYPE');
        }
        if (['a', 'button', 'select'].includes(tag) || ['button', 'checkbox', 'select'].includes(role)
            || hasEvent('click') || hasEvent('mousedown') || hasEvent('mouseup') || hasEvent('dblclick')
            || style.cursor === 'pointer'
            || e.hasAttribute('aria-haspopup')
            || (tag === 'input' && clickableInputs.includes(e.getAttribute('type')))
            || (tag === 'label' && document.getElementById(e.getAttribute('for')))
        ) {
            evts.push('CLICK');
        }
        return evts;
    }

    const results = {};
    function traverse(node, xpath) {
        if (node.nodeType === Node.ELEMENT_NODE) {
            const interactions = getInteractions(node);
            if (interactions.length > 0) {
                results[xpath] = interactions;
            }
        }
        const countByTag = {};
        for (let child = node.firstChild; child; child = child.nextSibling) {
            let tag = child.nodeName.toLowerCase();
            if (tag.includes(":")) continue; //namespace
            let isLocal = ['svg'].includes(tag);
            if (isLocal) {
                tag = `*[local-name() = '${tag}']`;
            }
            countByTag[tag] = (countByTag[tag] || 0) + 1;
            let childXpath = xpath + '/' + tag;
            if (countByTag[tag] > 1) {
                childXpath += '[' + countByTag[tag] + ']';
            }
            if (tag === 'iframe') {
                try {
                    traverse(child.contentWindow.document.body, childXpath + '/html/body');
                } catch (e) {
                    console.warn("iframe access blocked", child, e);
                }
            } else if (!isLocal) {
                traverse(child, childXpath);
            } 
        }
    }
    traverse(document.body, '/html/body');
    return results;
})();
"""

JS_GET_INTERACTIVES_IN_VIEWPORT = (
    """
const windowHeight = (window.innerHeight || document.documentElement.clientHeight);
const windowWidth = (window.innerWidth || document.documentElement.clientWidth);
return Object.fromEntries(Object.entries("""
    + js_wrap_function_call(JS_GET_INTERACTIVES)
    + """).filter(([xpath, evts]) => {
    const element = document.evaluate(xpath, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue;
    if (!element) return false;
    const rect = element.getBoundingClientRect();
    let iframe = element.ownerDocument.defaultView.frameElement;
    while (iframe) {
        const iframeRect = iframe.getBoundingClientRect();
        rect.top += iframeRect.top;
        rect.left += iframeRect.left;
        rect.bottom += iframeRect.top;
        rect.right += iframeRect.left;
        iframe = iframe.ownerDocument.defaultView.frameElement;
    }
    const elemCenter = {
        x: rect.left + element.offsetWidth / 2,
        y: rect.top + element.offsetHeight / 2
    };
    if (elemCenter.x < 0) return false;
    if (elemCenter.x > windowWidth) return false;
    if (elemCenter.y < 0) return false;
    if (elemCenter.y > windowHeight) return false;
    if (arguments?.[0] !== true) return true; // whenever to check for elements above
    let pointContainer = document.elementFromPoint(elemCenter.x, elemCenter.y);
    do {
        if (pointContainer === element) return true;
        if (pointContainer == null) return true;
    } while (pointContainer = pointContainer.parentNode);
    return false;
}));
"""
)

JS_WAIT_DOM_IDLE = """
return new Promise(resolve => {
    const timeout = arguments[0] || 10000;
    const stabilityThreshold = arguments[1] || 100;

    let mutationObserver;
    let timeoutId = null;

    const waitForIdle = () => {
        if (timeoutId) clearTimeout(timeoutId);
        timeoutId = setTimeout(() => resolve(true), stabilityThreshold);
    };
    mutationObserver = new MutationObserver(waitForIdle);
    mutationObserver.observe(document.body, {
        childList: true,
        attributes: true,
        subtree: true,
    });
    waitForIdle();

    setTimeout(() => {
        resolve(false);
        mutationObserver.disconnect();
        mutationObserver = null;
        if (timeoutId) {
            clearTimeout(timeoutId);
            timeoutId = null;
        }
    }, timeout);
});
"""



from io import BytesIO
import logging
import time
from typing import Any, List, Optional
from lavague.core.action_template import ActionTemplate
from lavague.core.context import Context, get_default_context
from lavague.core.exceptions import NavigationException
from lavague.core.extractors import (
    BaseExtractor,
    YamlFromMarkdownExtractor,
    DynamicExtractor,
)
from lavague.core.retrievers import BaseHtmlRetriever, get_default_retriever
from lavague.core.utilities.web_utils import (
    display_screenshot,
    sort_files_by_creation,
)
from lavague.core.exceptions import HallucinatedException
from lavague.core.logger import AgentLogger
from lavague.core.base_engine import BaseEngine, ActionResult
from lavague.core.base_driver import BaseDriver
from llama_index.core import QueryBundle, PromptTemplate
from PIL import Image
from llama_index.core.base.llms.base import BaseLLM
from llama_index.core.embeddings import BaseEmbedding

NAVIGATION_ENGINE_PROMPT_TEMPLATE = ActionTemplate(
    """
{driver_capability}

Here is a the next example to answer:

HTML:
{context_str}
Query: {query_str}
Completion:

""",
    YamlFromMarkdownExtractor(),
)

# JSON schema for the action shape
JSON_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "action": {
                "type": "object",
                "properties": {"name": {"type": "string"}, "args": {"type": "object"}},
                "required": ["name", "args"],
            }
        },
        "required": ["action"],
    },
}


logging_print = logging.getLogger(__name__)
logging_print.setLevel(logging.INFO)
format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(format)
logging_print.addHandler(ch)
logging_print.propagate = False


class NavigationEngine(BaseEngine):
    """
    NavigationEngine leverages the llm model and the to output code from the prompt and the html page.

    Args:
        driver (`BaseDriver`):
            The Web driver used to interact with the headless browser
        llm (`BaseLLM`)
            llama-index LLM that will generate the action
        retriever (`BaseHtmlRetriever`)
            Specify which algorithm will be used for RAG
        prompt_template (`PromptTemplate`)
            Squelette of the final prompt
        extractor (`BaseExtractor`)
            Specify how to extract the final code from the llm answer
        time_between_actions (`float`)
            Time between each action
        logger: (`AgentLogger`)
            Logger to log the actions taken by the agent
        embedding: (`BaseEmbedding`)
            Embedding to use for the retriever
    """

    def __init__(
        self,
        driver: BaseDriver,
        llm: BaseLLM = None,
        retriever: BaseHtmlRetriever = None,
        prompt_template: PromptTemplate = NAVIGATION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = DynamicExtractor(),
        time_between_actions: float = 1.5,
        n_attempts: int = 5,
        logger: AgentLogger = None,
        display: bool = False,
        raise_on_error: bool = False,
        embedding: BaseEmbedding = None,
    ):
        if llm is None:
            llm: BaseLLM = get_default_context().llm
        if retriever is None:
            retriever = get_default_retriever(driver, embedding=embedding)
        self.driver: BaseDriver = driver
        self.llm: BaseLLM = llm
        self.retriever: BaseHtmlRetriever = retriever
        self.prompt_template: PromptTemplate = prompt_template.partial_format(
            driver_capability=driver.get_capability()
        )
        self.extractor: BaseExtractor = extractor
        self.time_between_actions = time_between_actions
        self.logger = logger
        self.n_attempts = n_attempts
        self.display = display
        self.raise_on_error = raise_on_error
        self.shape_validator = JSON_SCHEMA

    @classmethod
    def from_context(
        cls,
        context: Context,
        driver: BaseDriver,
        retriever: BaseHtmlRetriever = None,
        prompt_template: PromptTemplate = NAVIGATION_ENGINE_PROMPT_TEMPLATE.prompt_template,
        extractor: BaseExtractor = DynamicExtractor(),
    ) -> "NavigationEngine":
        """
        Create an NavigationEngine from a context
        """
        return cls(
            driver,
            context.llm,
            retriever,
            prompt_template,
            extractor,
        )

    def get_nodes(self, query: str) -> List[str]:
        """
        Get the nodes from the html page

        Args:
            query (`str`): The query to search for

        Return:
            `List[str]`: The nodes
        """
        viewport_only = not self.driver.previously_scanned
        source_nodes = self.retriever.retrieve(
            QueryBundle(query_str=query), [self.driver.get_html()], viewport_only
        )
        return source_nodes

    def add_knowledge(self, knowledge: str):
        self.prompt_template = self.prompt_template + knowledge

    def get_action_from_context(self, context: str, query: str) -> str:
        """
        Generate the code from a query and a context
        """
        prompt = self.prompt_template.format(context_str=context, query_str=query)
        response = self.llm.complete(prompt).text
        code = self.extractor.extract(response)
        return code

    def set_display(self, display: bool):
        self.display = display

    def get_action(self, query: str) -> Optional[str]:
        """
        Generate the code from a query
        Args:
            query (`str`): Instructions given at the end of the prompt to tell the model what to do on the html page
        Return:
            `str`: The generated code
        """
        nodes = self.get_nodes(query)
        context = "\n".join(nodes)
        return self.get_action_from_context(context, query)

    def execute_instruction_gradio(self, instruction: str, action_engine: Any):
        """
        Generates code and executes it to answer the instruction

        Args:
            instruction (`str`): The instruction to perform

        Return:
            `bool`: True if the code was executed without error
            `Any`: The output of navigation is always None
        """

        from gradio import ChatMessage

        success = False
        action_full = ""
        output = None

        action_nb = 0
        navigation_log_total = []

        logging_print.debug("Query for retriever: " + instruction)

        start = time.time()
        source_nodes = self.get_nodes(instruction)
        end = time.time()
        retrieval_time = end - start

        llm_context = "\n".join(source_nodes)
        success = False
        logger = self.logger

        navigation_log = {
            "navigation_engine_input": instruction,
            "retrieved_html": source_nodes,
            "retrieval_time": retrieval_time,
            "retrieval_name": self.retriever.__class__.__name__,
        }

        action_outcomes = []
        for _ in range(self.n_attempts):
            if success:
                break
            if self.display:
                try:
                    scr_path = self.driver.get_current_screenshot_folder()
                    lst = sort_files_by_creation(scr_path)
                    for scr in lst:
                        img = Image.open(scr_path.as_posix() + "/" + scr)
                        display_screenshot(img)
                        time.sleep(0.35)
                except:
                    pass
            start = time.time()
            prompt = self.prompt_template.format(
                context_str=llm_context, query_str=instruction
            )
            response = self.llm.complete(prompt).text
            end = time.time()
            action_generation_time = end - start
            action_outcome = {
                "llm_raw_response": response,
                "action_generation_time": action_generation_time,
                "navigation_engine_full_prompt": prompt,
                "navigation_engine_llm": get_model_name(self.llm),
            }

            try:
                # We extract the action
                action = self.extractor.extract(response)
                self._verify_llm_reponse(response, llm_context)
                action_outcome["action"] = action
                action_full += action

                # Get information to see which elements are selected
                vision_data = self.driver.get_highlighted_element(action)
                action_full += action
                for item in vision_data:
                    screenshot = item["screenshot"]
                    if action_engine.screenshot_ratio != 1:
                        screenshot = screenshot.resize(
                            (
                                int(screenshot.width / action_engine.screenshot_ratio),
                                int(screenshot.height / action_engine.screenshot_ratio),
                            )
                        )
                    self.image_display = screenshot
                    yield (
                        self.objective,
                        self.url_input,
                        screenshot,
                        self.history,
                        output,
                    )

                self.driver.exec_code(action)
                self.history[-1] = ChatMessage(
                    role="assistant",
                    content=f"{action_engine.world_model_output}\n",
                    metadata={
                        "title": f"✅ Step {action_engine.curr_step} - {action_engine.curr_instruction}"
                    },
                )
                self.history.append(
                    ChatMessage(role="assistant", content="⏳ Loading the page...")
                )
                yield (
                    self.objective,
                    self.url_input,
                    self.image_display,
                    self.history,
                    output,
                )
                time.sleep(1)
                img = self.driver.get_screenshot_as_png()
                img = BytesIO(img)
                img = Image.open(img)
                if action_engine.screenshot_ratio != 1:
                    img = img.resize(
                        (
                            int(img.width / action_engine.screenshot_ratio),
                            int(img.height / action_engine.screenshot_ratio),
                        )
                    )
                self.image_display = img
                yield (
                    self.objective,
                    self.url_input,
                    self.image_display,
                    self.history,
                    output,
                )

                time.sleep(self.time_between_actions)

                success = True
                action_outcome["success"] = True
                navigation_log["vision_data"] = vision_data
            except Exception as e:
                logging_print.error(f"Navigation error: {e}")
                action_outcome["success"] = False
                action_outcome["error"] = str(e)
                if self.raise_on_error:
                    raise e

            action_outcomes.append(action_outcome)
            self.driver.wait_for_idle()

        navigation_log["action_outcomes"] = action_outcomes
        navigation_log["action_nb"] = action_nb
        action_nb += 1
        navigation_log_total.append(navigation_log)

        if not success:
            self.history[-1] = ChatMessage(
                role="assistant",
                content=f"{action_engine.world_model_output}",
                metadata={
                    "title": f"❌ Step {action_engine.curr_step + 1} - {action_engine.curr_instruction}"
                },
            )
        if logger:
            log = {
                "engine": "Navigation Engine",
                "instruction": instruction,
                "engine_log": navigation_log_total,
                "success": success,
                "output": None,
                "code": action_full,
            }

            logger.add_log(log)

        output = ActionResult(
            instruction=instruction,
            code=action_full,
            success=success,
            output=None,
        )
        action_engine.ret = output

        yield (
            self.objective,
            self.url_input,
            self.image_display,
            self.history,
            output.output,
        )

    def _verify_llm_reponse(self, llm_response: str, llm_context: str):
        """Make sure the action is performed on a given context to avoid hallucination"""
        actions_obj = self.extractor.extract_as_object(llm_response)
        if not isinstance(actions_obj, list):
            actions_obj = [actions_obj]

        for action_list in actions_obj:
            for action in action_list.get("actions", []):
                xpath = action.get("action", {}).get("args", {}).get("xpath", "")
                if xpath and xpath not in llm_context:
                    raise HallucinatedException(xpath)

    def execute_instruction(self, instruction: str) -> ActionResult:
        """
        Generates code and executes it to answer the instruction

        Args:
            instruction (`str`): The instruction to perform

        Return:
            `bool`: True if the code was executed without error
            `Any`: The output of navigation is always None
        """

        success = False
        action_full = ""
        action_nb = 0
        navigation_log_total = []

        logging_print.debug("Query for retriever: " + instruction)

        start = time.time()
        source_nodes = self.get_nodes(instruction)
        end = time.time()
        retrieval_time = end - start

        llm_context = "\n".join(source_nodes)
        success = False
        logger = self.logger

        navigation_log = {
            "navigation_engine_input": instruction,
            "retrieved_html": source_nodes,
            "retrieval_time": retrieval_time,
            "retrieval_name": self.retriever.__class__.__name__,
        }

        action_outcomes = []
        for _ in range(self.n_attempts):
            if success:
                break
            if self.display:
                try:
                    scr_path = self.driver.get_current_screenshot_folder()
                    lst = sort_files_by_creation(scr_path)
                    for scr in lst:
                        img = Image.open(scr_path.as_posix() + "/" + scr)
                        display_screenshot(img)
                        time.sleep(0.35)
                except:
                    pass
            start = time.time()
            prompt = self.prompt_template.format(
                context_str=llm_context, query_str=instruction
            )
            response = self.llm.complete(prompt).text
            end = time.time()
            action_generation_time = end - start
            action_outcome = {
                "llm_raw_response": response,
                "action_generation_time": action_generation_time,
                "navigation_engine_full_prompt": prompt,
                "navigation_engine_llm": get_model_name(self.llm),
            }

            try:
                # We extract the action
                action = self.extractor.extract(response)
                self._verify_llm_reponse(response, llm_context)

                action_outcome["action"] = action
                action_full += action

                # Get information to see which elements are selected
                vision_data = self.driver.get_highlighted_element(action)
                if self.display:
                    for item in vision_data:
                        display_screenshot(item["screenshot"])
                        time.sleep(0.2)
                self.driver.exec_code(action)
                time.sleep(self.time_between_actions)
                if self.display:
                    try:
                        screenshot = self.driver.get_screenshot_as_png()
                        screenshot = BytesIO(screenshot)
                        screenshot = Image.open(screenshot)
                        display_screenshot(screenshot)
                    except:
                        pass
                success = True
                action_outcome["success"] = True
                navigation_log["vision_data"] = vision_data
            except Exception as e:
                logging_print.error(f"Navigation error: {e}")
                action_outcome["success"] = False
                action_outcome["error"] = str(e)
                if self.raise_on_error:
                    raise e

            action_outcomes.append(action_outcome)

        navigation_log["action_outcomes"] = action_outcomes
        navigation_log["action_nb"] = action_nb
        action_nb += 1
        navigation_log_total.append(navigation_log)

        if logger:
            log = {
                "engine": "Navigation Engine",
                "instruction": instruction,
                "engine_log": navigation_log_total,
                "success": success,
                "output": None,
                "code": action_full,
            }

            logger.add_log(log)

        return ActionResult(
            instruction=instruction,
            code=action_full,
            success=success,
            output=None,
        )


class NavigationControl(BaseEngine):
    driver: BaseDriver
    time_between_actions: float
    logger: AgentLogger

    def __init__(
        self,
        driver: BaseDriver,
        time_between_actions: float = 1.5,
        logger: AgentLogger = None,
        navigation_engine: Optional[NavigationEngine] = None,
    ) -> None:
        self.driver: BaseDriver = driver
        self.time_between_actions = time_between_actions
        self.logger = logger
        self.display = False
        self.navigation_engine = navigation_engine

    def set_display(self, display: bool):
        self.display = display

    def execute_instruction(self, instruction: str) -> ActionResult:
        import inspect

        code = ""
        output = None
        success = True
        logger = self.logger

        try:
            if "SCROLL_DOWN" in instruction:
                self.driver.scroll_down()
                code = inspect.getsource(self.driver.scroll_down)
            elif "SCROLL_UP" in instruction:
                self.driver.scroll_up()
                code = inspect.getsource(self.driver.scroll_up)
            elif "WAIT" in instruction:
                self.driver.wait(self.time_between_actions)
                code = inspect.getsource(self.driver.wait)
            elif "BACK" in instruction:
                self.driver.back()
                code = inspect.getsource(self.driver.back)
            elif "SCAN" in instruction:
                self.driver.get_screenshots_whole_page()
                code = inspect.getsource(self.driver.get_screenshots_whole_page)
            elif "MAXIMIZE_WINDOW" in instruction:
                self.driver.maximize_window()
                code = inspect.getsource(self.driver.maximize_window)
            elif "SWITCH_TAB" in instruction:
                tab_id = int(instruction.split(" ")[1])
                try:
                    self.driver.switch_tab(tab_id=tab_id)
                except Exception as e:
                    raise ValueError(f"Error while switching tab: {e}")
                code = inspect.getsource(self.driver.switch_tab)
            else:
                raise ValueError(f"Unknown instruction: {instruction}")

        except NavigationException as e:
            success = False
            output = str(e)
            logging_print.error(f"Navigation error: {e}")

        if logger:
            log = {
                "engine": "Navigation Controls",
                "instruction": instruction,
                "engine_log": None,
                "success": success,
                "output": output,
                "code": code,
            }
            logger.add_log(log)

        self.driver.wait_for_idle()

        return ActionResult(
            instruction=instruction, code=code, success=success, output=output
        )


def get_model_name(llm: BaseLLM) -> str:
    try:
        # Try accessing the 'model' attribute
        return llm.model
    except AttributeError:
        try:
            # Try accessing the 'model_name' attribute
            return llm.model_name
        except AttributeError:
            return "Unknown"



"""TODO: Make this class generalizable"""

from lavague.core.logger import AgentLogger, Loggable


class ShortTermMemory(Loggable):
    """
    Short term memory module of the agent.
    """

    def __init__(self, user_data=None, logger: AgentLogger = None) -> None:
        current_state = {
            "external_observations": {
                "vision": "[SCREENSHOT]",
            },
            "internal_state": {
                "user_inputs": [],
                "agent_outputs": [],
            },
        }
        self.logger = logger

        if user_data:
            current_state["internal_state"]["user_inputs"].append(user_data)

        self.current_state = current_state

        self.previous_instructions: str = "[NONE]"
        self.last_engine: str = "[NONE]"

    def set_user_data(self, user_data=None):
        if user_data:
            self.current_state["internal_state"]["user_inputs"].append(user_data)

    def get_state(self):
        current_state = self.current_state
        past = {
            "previous_instructions": str(self.previous_instructions),
            "last_engine": self.last_engine,
        }

        logger = self.logger

        if logger:
            log = {
                "current_state": current_state,
                "past": past,
            }
            logger.add_log(log)
        return current_state, past

    def update_state(self, instruction, last_engine, success, output):
        if not success:
            instruction = "[FAILED] " + instruction

        if output:
            self.current_state["internal_state"]["agent_outputs"].append(output)

        if self.previous_instructions == "[NONE]":
            self.previous_instructions = f"""
- {instruction}"""
        else:
            self.previous_instructions += f"""
- {instruction}"""
        self.last_engine = last_engine



import requests
import pkg_resources
import warnings

YELLOW = "\033[93m"
RESET = "\033[0m"


def compare_versions(version1, version2):
    v1 = list(map(int, version1.split(".")))
    v2 = list(map(int, version2.split(".")))

    while len(v1) < len(v2):
        v1.append(0)
    while len(v2) < len(v1):
        v2.append(0)

    for i in range(len(v1)):
        if v1[i] > v2[i]:
            return 1
        elif v1[i] < v2[i]:
            return -1

    return 0


def get_installed_version(dist_name, lookup_dirs=None):
    req = pkg_resources.Requirement.parse(dist_name)

    if lookup_dirs is None:
        working_set = pkg_resources.WorkingSet()
    else:
        working_set = pkg_resources.WorkingSet(lookup_dirs)

    dist = working_set.find(req)
    return dist.version if dist else None


def check_latest_version():
    package_version = get_installed_version("lavague-core")
    url = "https://pypi.org/pypi/lavague-core/json"
    response = requests.get(url)
    data = response.json()
    latest_version = data["info"]["version"]
    if compare_versions(package_version, latest_version) < 0:
        warnings.warn(
            YELLOW
            + f"You are using lavague-core version {package_version}, however version {latest_version} is "
            "available. You should consider upgrading via the "
            "'pip install --upgrade lavague-core' command." + RESET,
            UserWarning,
        )



from lavague.core.python_engine import PythonEngine
from lavague.core.context import Context, get_default_context
from lavague.core.extractors import (
    PythonFromMarkdownExtractor,
    JsonFromMarkdownExtractor,
)
from lavague.core.retrievers import OpsmSplitRetriever
from lavague.core.world_model import WorldModel
from lavague.core.utilities.version_checker import check_latest_version
from lavague.core.action_engine import ActionEngine
from lavague.core.agents import WebAgent

import os
import warnings


def telemetry_warning():
    telemetry_var = os.getenv("LAVAGUE_TELEMETRY")
    if telemetry_var != "NONE":
        warning_message = "\033[93mTelemetry is turned on. To turn off telemetry, set your LAVAGUE_TELEMETRY to 'NONE'\033[0m"
        warnings.warn(warning_message, UserWarning)


telemetry_warning()
try:
    check_latest_version()
except:
    pass



from abc import ABC, abstractmethod
from typing import Any, Optional
from lavague.core.display import Display
from lavague.core.logger import Loggable
from dataclasses import dataclass


@dataclass
class ActionResult:
    """Represent the result of executing an instruction"""

    instruction: str
    code: str
    success: bool
    output: Any
    total_estimated_tokens: Optional[int] = 0
    total_estimated_cost: Optional[float] = 0


class BaseEngine(ABC, Loggable, Display):
    @abstractmethod
    def execute_instruction(self, instruction: str) -> ActionResult:
        pass



from typing import Any


class Display:
    display: bool = False
    gradio_mode: bool = False
    image_display: Any = None
    objective: Any = None
    url_input: Any = None
    history: Any = None

    def set_display(self, display: bool):
        self.display = display

    def set_gradio_mode(
        self,
        gradio_mode: bool,
        objective,
        url_input,
        image_display,
        history,
    ):
        self.gradio_mode = gradio_mode
        self.image_display = image_display
        self.objective = objective
        self.url_input = url_input
        self.history = history



import yaml
import os


def load_yaml_file(file_path):
    try:
        with open(file_path, "r") as file:
            data = yaml.safe_load(file)
        return data
    except FileNotFoundError:
        print(f"Error: The pricing config file, {file_path} was not found.")
        return None
    except yaml.YAMLError as exc:
        print(f"Error parsing YAML file: {exc}")
        return None


def get_pricing_data():
    pricing_file = "pricing_config.yml"
    pricing_file_path = os.path.join(os.path.dirname(__file__), pricing_file)

    data = load_yaml_file(pricing_file_path)
    models = data.get("models", {})
    return models


def build_summary_table(token_summary: dict, verbose: bool = True):
    # calculate totals we don't have from the logs
    total_input = (
        token_summary["world_model_input_tokens"]
        + token_summary["action_engine_input_tokens"]
        + token_summary["total_embedding_tokens"]
    )
    total_output = (
        token_summary["world_model_output_tokens"]
        + token_summary["action_engine_output_tokens"]
    )

    # build table (header)
    header = f"{'Component':<16} | {'Input':<10} | {'Output':<10} | {'Total':<10} | {'Cost (USD)':<10} |\n"
    line = "-" * 70 + "\n"

    world_model_row, action_engine_row, embeddings_row = "", "", ""
    if verbose:
        # build table (rows)
        world_model_row = f"{'World Model':<16} | {token_summary['world_model_input_tokens']:<10} | {token_summary['world_model_output_tokens']:<10} | {token_summary['total_world_model_tokens']:<10} | $ {token_summary['total_world_model_cost']:<8.4f} |\n"
        action_engine_row = f"{'Action Engine':<16} | {token_summary['action_engine_input_tokens']:<10} | {token_summary['action_engine_output_tokens']:<10} | {token_summary['total_action_engine_tokens']:<10} | $ {token_summary['total_action_engine_cost']:<8.4f} |\n"
        embeddings_row = f"{'Embeddings':<16} | {token_summary['total_embedding_tokens']:<10} | {' ':<10} | {token_summary['total_embedding_tokens']:<10} | $ {token_summary['total_embedding_cost']:<8.4f} |\n"
        embeddings_row += line
    total_row = f"{'Total':<16} | {total_input:<10} | {total_output:<10} | {token_summary['total_step_tokens']:<10} | $ {token_summary['total_step_cost']:<8.4f} |\n"

    # combine table
    table = (
        "\n"
        + header
        + line
        + world_model_row
        + action_engine_row
        + embeddings_row
        + total_row
    )

    return table



import os
from time import sleep
import random

term_size = 0
icons = ["  ", "🌊", "🏄", "🏄", "🏄", "🏄", "🏄", "🐬", "🦈", "🦀", "🐙", "🐟"]


def _full():
    frames = []
    waves = []
    bg = [0] * term_size
    i = 10
    iend = term_size - 1
    while iend > 0:
        line = list(bg)
        if i % 10 == 0:
            waves.append(0)
        for iw in range(len(waves)):
            line[waves[iw]] = 1
            if waves[iw] >= iend:
                continue
            waves[iw] += 1
            if waves[iw] == iend:
                iend -= 1
        frames.append(line)
        i += 1
    return frames


def _surfing():
    frames = []
    bg = [1] * term_size
    rider = random.randint(2, len(icons) - 1)
    for i in reversed(range(term_size)):
        line = list(bg)
        line[i] = rider
        frames.append(line)
    frames.append(bg)
    return frames


def _display(frames, sleep_time):
    for line in frames:
        print("".join(map(lambda x: icons[x], line)), end="\r", flush=True)
        sleep(sleep_time)


def clear_animation():
    print("  " * term_size)


def lavague_unicode_animation():
    global term_size
    term_size = os.get_terminal_size().columns // 2
    _display(_full(), 0.1)
    while True:
        _display(_surfing(), 0.1)
        sleep(random.uniform(0.5, 5))



from io import BytesIO
import logging
import os
import shutil
from typing import Any, Optional

from lavague.core.action_engine import ActionEngine
from lavague.core.world_model import WorldModel
from lavague.core.utilities.format_utils import (
    extract_before_next_engine,
    extract_next_engine,
    extract_world_model_instruction,
    replace_hyphens,
)
from lavague.core.logger import AgentLogger, LocalDBLogger
from lavague.core.memory import ShortTermMemory
from lavague.core.base_driver import BaseDriver
from lavague.core.base_engine import ActionResult
from lavague.core.utilities.telemetry import send_telemetry
from PIL import Image
from IPython.display import display, HTML, Code
from lavague.core.token_counter import TokenCounter
from lavague.core.utilities.config import is_flag_true

logging_print = logging.getLogger(__name__)
logging_print.setLevel(logging.INFO)
format = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(format)
logging_print.addHandler(ch)
logging_print.propagate = False


class WebAgent:
    """
    Web agent class, for now only works with selenium.
    """

    def __init__(
        self,
        world_model: WorldModel,
        action_engine: ActionEngine,
        token_counter: Optional[TokenCounter] = None,
        n_steps: int = 10,
        clean_screenshot_folder: bool = True,
        logger: AgentLogger = None,
    ):
        self.driver: BaseDriver = action_engine.driver
        self.action_engine: ActionEngine = action_engine
        self.world_model: WorldModel = world_model
        self.st_memory = ShortTermMemory()
        self.token_counter = token_counter
        self.interrupted = False

        self.n_steps = n_steps

        self.output = ""

        self.clean_screenshot_folder = clean_screenshot_folder

        if logger is None:
            self.logger: AgentLogger = AgentLogger()
        else:
            self.logger = logger

        self.action_engine.set_logger_all(self.logger)
        self.world_model.set_logger(self.logger)
        self.st_memory.set_logger(self.logger)

        if self.clean_screenshot_folder:
            try:
                if os.path.isdir("screenshots"):
                    shutil.rmtree("screenshots")
                logging_print.info("Screenshot folder cleared")
            except:
                pass

        self.result = ActionResult(
            instruction=None,
            code=self.driver.code_for_init(),
            success=False,
            output=None,
            total_estimated_tokens=0,
            total_estimated_cost=0.0,
        )

    def get(self, url):
        self.driver.get(url)
        self.driver.wait_for_idle()
        self.result.code += self.driver.code_for_get(url) + "\n"

    def demo(
        self,
        objective: str = "",
        user_data=None,
        screenshot_ratio: float = 1,
    ):
        try:
            from lavague.gradio import GradioAgentDemo

            grad = GradioAgentDemo(objective, None, self, user_data, screenshot_ratio)
            grad.launch()
        except ImportError:
            raise ImportError(
                "`lavague-gradio` package not found, "
                "please run `pip install lavague-gradio`"
            )

    def _run_demo(
        self,
        objective: str,
        user_data=None,
        display: bool = False,
        objective_obj: Any = None,
        url_input: Any = None,
        image_display: Any = None,
        history: Any = None,
        screenshot_ratio: float = 1,
    ):
        """Internal run method for the gradio demo. Do not use directly. Use run instead."""

        from gradio import ChatMessage

        driver: BaseDriver = self.driver
        logger = self.logger
        n_steps = self.n_steps
        self.action_engine.set_display_all(display)
        output = None
        success = True

        try:
            if os.path.isdir("screenshots"):
                shutil.rmtree("screenshots")
            logging_print.info("Screenshot folder cleared")
        except:
            pass

        self.st_memory = ShortTermMemory()
        st_memory = self.st_memory
        world_model = self.world_model

        if user_data:
            st_memory.set_user_data(user_data)

        obs = driver.get_obs()

        logger.clear_logs()
        for curr_step in range(n_steps):
            current_state, past = st_memory.get_state()

            world_model_output = world_model.get_instruction(
                objective, current_state, past, obs
            )
            self.action_engine.world_model_output = replace_hyphens(
                extract_before_next_engine(world_model_output)
            )
            logging_print.info(world_model_output)
            next_engine_name = extract_next_engine(world_model_output)
            instruction = extract_world_model_instruction(world_model_output)

            self.action_engine.screenshot_ratio = screenshot_ratio
            img = self.driver.get_screenshot_as_png()
            img = BytesIO(img)
            img = Image.open(img)
            if screenshot_ratio != 1:
                img = img.resize(
                    (
                        int(img.width / screenshot_ratio),
                        int(img.height / screenshot_ratio),
                    )
                )
            image_display = img

            self.action_engine.curr_step = curr_step + 1
            self.action_engine.curr_instruction = instruction

            if instruction.find("[NONE]") == -1 and next_engine_name != "COMPLETE":
                history[-1] = ChatMessage(
                    role="assistant",
                    content=f"{self.action_engine.world_model_output}",
                    metadata={
                        "title": f"⏳ Step {curr_step + 1} - {self.action_engine.curr_instruction}"
                    },
                )
            yield (
                objective_obj,
                url_input,
                image_display,
                history,
                output,
            )

            if next_engine_name == "COMPLETE" or next_engine_name == "SUCCESS":
                output = instruction
                logging_print.info("Objective reached. Stopping...")

                logger.add_log(obs)
                logger.end_step()
                break

            yield from self.action_engine.dispatch_instruction_gradio(
                next_engine_name, instruction
            )

            success = self.action_engine.ret.success
            output = self.action_engine.ret.output

            st_memory.update_state(
                instruction,
                next_engine_name,
                success,
                output,
            )

            logger.add_log(obs)
            logger.end_step()

            obs = driver.get_obs()
            if next_engine_name != "Navigation Engine":
                if success:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"{self.action_engine.world_model_output}",
                        metadata={"title": f"✅ Step {curr_step + 1} - {instruction}"},
                    )
                else:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"{self.action_engine.world_model_output}",
                        metadata={"title": f"❌ Step {curr_step + 1} - {instruction}"},
                    )
                history.append(
                    ChatMessage(
                        role="assistant", content="⏳ Thinking of next steps..."
                    )
                )
            else:
                history[-1] = ChatMessage(
                    role="assistant", content=f"⏳ Thinking of next steps..."
                )
            url_input = self.action_engine.driver.get_url()
            img = self.driver.get_screenshot_as_png()
            img = BytesIO(img)
            img = Image.open(img)
            if screenshot_ratio != 1:
                img = img.resize(
                    (
                        int(img.width / screenshot_ratio),
                        int(img.height / screenshot_ratio),
                    )
                )
            image_display = img
            yield (
                objective_obj,
                url_input,
                image_display,
                history,
                output,
            )
        send_telemetry(logger.return_pandas(), origin="gradio")
        url_input = self.action_engine.driver.get_url()
        if output is not None:
            if len(output) > 0 and output.strip() != "[NONE]":
                history[-1] = ChatMessage(
                    role="assistant", content=output, metadata={"title": f"🌊 Output"}
                )
            elif len(output) == 0 or output.strip() == "[NONE]":
                if success:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"The objective was successfully executed after {curr_step} step(s).",
                        metadata={"title": f"🌊 Objective reached"},
                    )
                else:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"The objective was not successfully executed after {curr_step} step(s).",
                        metadata={"title": f"❌ Failed to reach objective"},
                    )
            else:
                if success:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"The objective was successfully executed after {curr_step} step(s).",
                        metadata={"title": f"🌊 Objective reached"},
                    )
                else:
                    history[-1] = ChatMessage(
                        role="assistant",
                        content=f"The objective was not successfully executed after {curr_step} step(s).",
                        metadata={"title": f"❌ Failed to reach objective"},
                    )

        yield (
            objective_obj,
            url_input,
            image_display,
            history,
            output,
        )

    def run_step(self, objective: str) -> Optional[ActionResult]:
        obs = self.driver.get_obs()
        current_state, past = self.st_memory.get_state()

        world_model_output = self.world_model.get_instruction(
            objective, current_state, past, obs
        )
        logging_print.info(world_model_output)
        next_engine_name = extract_next_engine(world_model_output)
        instruction = extract_world_model_instruction(world_model_output)

        if next_engine_name == "COMPLETE" or next_engine_name == "SUCCESS":
            self.result.success = True
            self.result.output = instruction
            logging_print.info("Objective reached. Stopping...")
            self.logger.add_log(obs)

            self.process_token_usage()
            self.logger.end_step()
            return self.result

        action_result = self.action_engine.dispatch_instruction(
            next_engine_name, instruction
        )
        if action_result.success:
            self.result.code += action_result.code
            self.result.output = action_result.output
        self.st_memory.update_state(
            instruction,
            next_engine_name,
            action_result.success,
            action_result.output,
        )
        self.logger.add_log(obs)

        self.process_token_usage()
        self.logger.end_step()

    def prepare_run(self, display: bool = False, user_data=None):
        self.action_engine.set_display_all(display)
        if user_data:
            self.st_memory.set_user_data(user_data)
        self.logger.new_run()

    def run(
        self,
        objective: str,
        user_data=None,
        display: bool = False,
        log_to_db: bool = is_flag_true("LAVAGUE_LOG_TO_DB"),
        step_by_step=False,
    ) -> ActionResult:
        self.interrupted = False
        self.prepare_run(display=display, user_data=user_data)

        try:
            for _ in range(self.n_steps):
                result = self.run_step(objective)

                if result is not None:
                    break

                if step_by_step:
                    input("Press ENTER to continue")

        except KeyboardInterrupt:
            logging_print.warning("The agent was interrupted.")
            self.interrupted = True
            pass
        except Exception as e:
            logging_print.error(f"Error while running the agent: {e}")
            self.interrupted = True
            raise e
        finally:
            origin = self.origin if hasattr(self, "origin") else "lavague"
            send_telemetry(self.logger.return_pandas(), origin=origin)
            if log_to_db:
                local_db_logger = LocalDBLogger()
                local_db_logger.insert_logs(self)
        return self.result

    def process_token_usage(self):
        if self.token_counter is not None:
            token_counts, token_costs = self.token_counter.process_token_usage(
                self.world_model, self.action_engine, result_to_update=self.result
            )
            self.logger.add_log(token_counts)
            self.logger.add_log(token_costs)

    def display_previous_nodes(self, steps: int) -> None:
        """prints out all nodes per each sub-instruction for given steps"""
        dflogs = self.logger.return_pandas()
        # check if dflogs are not null and not empty and engine_log is present in dflogs columns
        if (
            dflogs is not None
            and dflogs.empty is False
            and "engine_log" in dflogs.columns
        ):
            if steps > len(dflogs):
                print(
                    f"Previous steps: {len(dflogs)}\nrequested steps: {steps}\nshowing available steps"
                )
            steps = len(dflogs) if steps > len(dflogs) else steps
            for step in range(steps):
                print(f"Step: {step}")
                sub_ins = 0
                if isinstance(dflogs.at[step, "engine_log"], list):
                    for subinst in dflogs.at[step, "engine_log"]:
                        print(f"Sub-Instruction: {sub_ins}")
                        sub_ins += 1
                        x = 0
                        for node in subinst["retrieved_html"]:
                            print(f"Node {x}")
                            x = x + 1
                            display(HTML(node))  # Display node as visual element
                            display(Code(node, language="html"))  # Display code
        else:
            print(
                f"No previous nodes available. Please run the agent atleast once to view previous steps"
            )

    def display_all_nodes(self) -> None:
        """prints out all nodes per each sub-instruction"""
        dflogs = self.logger.return_pandas()
        # check if dflogs are not null and not empty and engine_log is present in dflogs columns
        if (
            dflogs is not None
            and dflogs.empty is False
            and "engine_log" in dflogs.columns
        ):
            print(f"Number of steps: {len(dflogs)}")
            steps = len(dflogs)
            for step in range(steps):
                print(f"Step: {step}")
                sub_ins = 0
                if isinstance(dflogs.at[step, "engine_log"], list):
                    for subinst in dflogs.at[step, "engine_log"]:
                        print(f"Sub-Instruction: {sub_ins}")
                        sub_ins += 1
                        x = 0
                        for node in subinst["retrieved_html"]:
                            print(f"Node: {x}")
                            x = x + 1
                            display(HTML(node))  # Display node as visual element
                            display(Code(node, language="html"))  # Display code
        else:
            print(
                f"No previous nodes available. Please run the agent atleast once to view previous steps"
            )

    def set_origin(self, origin: str):
        self.origin = origin



from IPython.display import display, clear_output
from PIL.PngImagePlugin import PngImageFile
import base64
import os


def sort_files_by_creation(directory):
    def get_creation_time(item):
        item_path = os.path.join(directory, item)
        return os.path.getctime(item_path)

    items = os.listdir(directory)
    sorted_items = sorted(items, key=get_creation_time)
    return sorted_items


# Function to encode the image
def encode_image(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")


def display_screenshot(img: PngImageFile):
    clear_output()
    if img.mode == "RGBA":
        img = img.convert("RGB")

    try:
        __IPYTHON__
        display(img)
    except:
        img.show()



from llama_index.core import PromptTemplate
from lavague.core.extractors import BaseExtractor


class ActionTemplate:
    """
    Define a template, extractor pair
    """

    def __init__(self, prompt_template: str, extractor: BaseExtractor):
        self.prompt_template: PromptTemplate = PromptTemplate(prompt_template)
        self.extractor: BaseExtractor = extractor



import uuid
import pandas as pd
import os
from PIL import Image
import json
import sqlite3
import io
from pandas.core.frame import DataFrame


def load_images_from_folder(folder_path):
    images = []
    for filename in os.listdir(folder_path):
        if (
            filename.endswith(".png")
            or filename.endswith(".jpg")
            or filename.endswith(".jpeg")
        ):
            img_path = os.path.join(folder_path, filename)
            img = Image.open(img_path)
            images.append(img)
    return images


class AgentLogger:
    def __init__(self):
        self.logs = []

    def clear_logs(self):
        self.run_id = str(uuid.uuid4())
        self.logs = []
        self.current_row = {}
        self.current_step = 0

    def new_run(self):
        self.run_id = str(uuid.uuid4())
        self.current_row = {}
        self.current_step = 0

    def end_step(self):
        row = self.current_row
        row["run_id"] = self.run_id
        row["step"] = self.current_step
        self.logs.append(row)
        self.current_row = {}
        self.current_step += 1

    def add_log(self, log: dict):
        for k, v in log.items():
            self.current_row[k] = v

    def return_pandas(self) -> pd.DataFrame:
        df = pd.DataFrame(self.logs)
        return df


class LocalLogger(AgentLogger):
    def __init__(self, log_file_path: str, ignore_keys: list[str] = None):
        self.log_file_path = log_file_path
        self.ignore_keys = ignore_keys
        if self.ignore_keys is None:
            self.ignore_keys = ["screenshots", "html"]
        super().__init__()
        with open(self.log_file_path, "w") as f:
            f.write("")

    def clear_logs(self):
        super().clear_logs()
        # Clear the log file
        with open(self.log_file_path, "w") as f:
            f.write("")

    def add_log(self, log: dict):
        super().add_log(log)
        # Write the log to the file
        with open(self.log_file_path, "a") as f:
            f.write(self.serialize_dict(log))
            f.write("\n")  # Ensure each log entry is on a new line

    # Custom function to serialize non-serializable objects
    def custom_serializer(self, obj):
        if isinstance(obj, dict):
            return {
                k: self.custom_serializer(v)
                for k, v in obj.items()
                if k not in self.ignore_keys
            }
        elif isinstance(obj, list):
            return [self.custom_serializer(v) for v in obj]
        try:
            json.dumps(obj)
            return obj
        except (TypeError, OverflowError):
            return None

    # Function to serialize dictionary with ignoring non-serializable properties
    def serialize_dict(self, input_dict):
        return json.dumps(self.custom_serializer(input_dict))


class LocalDBLogger(AgentLogger):
    def __init__(self, db_name: str = "lavague_logs.db"):
        self.db_name = db_name
        self.ensure_connection()

    def ensure_connection(self):
        try:
            with sqlite3.connect(self.db_name) as conn:
                print("Connected to SQLite")
        except sqlite3.Error as error:
            print("Error occurred while connecting to SQLite -", error)

    def create_or_alter_table(self, df_logs: DataFrame):
        columns = df_logs.columns
        try:
            with sqlite3.connect(self.db_name) as conn:
                cursor = conn.cursor()

                # check if the table exists
                cursor.execute(
                    "SELECT name FROM sqlite_master WHERE type='table' AND name='Logs'"
                )
                table_exists = cursor.fetchone() is not None

                if not table_exists:
                    # create the table with all columns as TEXT
                    create_query = f"CREATE TABLE Logs ({', '.join([f'{col} TEXT' for col in columns])})"
                    cursor.execute(create_query)
                else:
                    # get existing columns
                    cursor.execute("PRAGMA table_info(Logs)")
                    existing_columns = set(row[1] for row in cursor.fetchall())

                    # add new columns
                    for col in columns:
                        if col not in existing_columns:
                            alter_query = f"ALTER TABLE Logs ADD COLUMN {col} TEXT"
                            cursor.execute(alter_query)

                conn.commit()
                print("Table created or altered successfully")
        except sqlite3.Error as error:
            print("Error occurred while creating or altering table -", error)

    def insert_logs(self, agent) -> None:
        if agent:
            df_logs = agent.logger.return_pandas()
            try:
                self.create_or_alter_table(df_logs)

                with sqlite3.connect(self.db_name) as conn:
                    cursor = conn.cursor()

                    columns = ", ".join(df_logs.columns)
                    placeholders = ", ".join(["?" for _ in df_logs.columns])
                    insert_query = (
                        f"INSERT INTO Logs ({columns}) VALUES ({placeholders})"
                    )

                    data_to_insert = self.format_df_logs_to_sqlite3_types(df_logs)
                    cursor.executemany(insert_query, data_to_insert)

                    conn.commit()
                    print("Log insert complete")
            except sqlite3.Error as error:
                print("Error occurred while inserting logs -", error)

    def format_df_logs_to_sqlite3_types(self, df_logs: DataFrame) -> list:
        if df_logs is not None and isinstance(df_logs, DataFrame):
            data = []
            for index, r in df_logs.iterrows():
                t = []
                for col in df_logs:
                    i = r[col]
                    if col == "screenshots":
                        t.append(str(self.convertImgToBlob(i)))
                    elif type(i) == str or type(i) == float or type(i) == int:
                        t.append(i)
                    else:
                        t.append(str(i))
                data.append(t)
            return data
        else:
            print("Please pass a dataframe!")

    def convertImgToBlob(self, imageList: list) -> list:
        blobList = []
        image_bytes = io.BytesIO()
        for img in imageList:
            img.save(image_bytes, format="PNG")
            blobList.append(image_bytes)
        return blobList


class Loggable:
    logger: AgentLogger

    def set_logger(self, logger: AgentLogger):
        self.logger = logger
